# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
from argparse import Namespace, _ActionsContainer
from collections import defaultdict

from pex.dist_metadata import Distribution, Requirement
from pex.orderedset import OrderedSet
from pex.pep_503 import ProjectName
from pex.pex_info import PexInfo
from pex.targets import Target
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import DefaultDict, Iterable, Mapping, Optional, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class DependencyConfiguration(object):
    @classmethod
    def create(
        cls,
        excluded=(),  # type: Iterable[Union[str, Requirement]]
        overridden=(),  # type: Iterable[Union[str, Requirement]]
    ):
        # type: (...) -> DependencyConfiguration

        overridden_projects = defaultdict(
            OrderedSet
        )  # type: DefaultDict[ProjectName, OrderedSet[Requirement]]
        for o in overridden:
            override = o if isinstance(o, Requirement) else Requirement.parse(o)
            overridden_projects[override.project_name].add(override)

        return cls(
            excluded=tuple(
                OrderedSet(
                    req if isinstance(req, Requirement) else Requirement.parse(req)
                    for req in excluded
                )
            ),
            overridden={
                project_name: tuple(overrides)
                for project_name, overrides in overridden_projects.items()
            },
        )

    @classmethod
    def from_pex_info(cls, pex_info):
        # type: (PexInfo) -> DependencyConfiguration
        return cls.create(pex_info.excluded, pex_info.overridden)

    excluded = attr.ib(default=())  # type: Tuple[Requirement, ...]
    overridden = attr.ib(factory=dict)  # type: Mapping[ProjectName, Tuple[Requirement, ...]]

    def configure(self, pex_info):
        # type: (PexInfo) -> None
        for excluded in self.excluded:
            pex_info.add_exclude(excluded)
        for override in sorted(
            itertools.chain.from_iterable(self.overridden.values()), key=lambda req: req.key
        ):
            pex_info.add_override(override)

    def excluded_by(self, item):
        # type: (Union[Distribution, Requirement]) -> Tuple[Requirement, ...]
        return tuple(req for req in self.excluded if item in req)

    def all_overrides(self):
        # type: () -> Tuple[Requirement, ...]
        return tuple(itertools.chain.from_iterable(self.overridden.values()))

    def overrides_for(self, requirement):
        # type: (Requirement) -> Tuple[Requirement, ...]
        return self.overridden.get(requirement.project_name, ())

    def overridden_by(self, requirement, target):
        # type: (Requirement, Target) -> Optional[Requirement]
        overrides = self.overrides_for(requirement)
        applicable_overrides = [
            override for override in overrides if target.requirement_applies(override)
        ]
        if len(applicable_overrides) > 1:
            raise ValueError(
                "Invalid override configuration for target {target}.\n"
                "More than one applicable override was found for {requirement}:\n"
                "{overrides}".format(
                    requirement=repr(str(requirement)),
                    target=target.render_description(),
                    overrides="\n".join(
                        "{index}. {override}".format(index=index, override=repr(str(override)))
                        for index, override in enumerate(applicable_overrides, start=1)
                    ),
                )
            )
        return applicable_overrides[0] if applicable_overrides else None

    def merge(self, other):
        # type: (DependencyConfiguration) -> DependencyConfiguration

        excluded = OrderedSet(self.excluded)
        excluded.update(other.excluded)

        overridden = defaultdict(
            OrderedSet
        )  # type: DefaultDict[ProjectName, OrderedSet[Requirement]]
        for project_name, overrides in itertools.chain(
            self.overridden.items(), other.overridden.items()
        ):
            overridden[project_name].update(overrides)

        return DependencyConfiguration(
            excluded=tuple(excluded),
            overridden={
                project_name: tuple(overrides) for project_name, overrides in overridden.items()
            },
        )


def register(parser):
    # type: (_ActionsContainer) -> None
    """Register dependency configuration options with the given parser.

    :param parser: The parser to register dependency configuration options with.
    """
    parser.add_argument(
        "--exclude",
        dest="excluded",
        default=[],
        type=str,
        action="append",
        help=(
            "Specifies a requirement to exclude from the built PEX. Any distribution included in "
            "the PEX's resolve that matches the requirement is excluded from the built PEX along "
            "with all of its transitive dependencies that are not also required by other "
            "non-excluded distributions. At runtime, the PEX will boot without checking the "
            "excluded dependencies are available (say, via `--inherit-path`). This option can be "
            "used multiple times."
        ),
    )
    parser.add_argument(
        "--override",
        dest="overridden",
        default=[],
        type=str,
        action="append",
        help=(
            "Specifies a transitive requirement to override when resolving. Any distribution "
            "requirement in the PEX's resolve that matches the override project name is replaced "
            "with the given override requirement. This option can be used multiple times."
        ),
    )


def configure(options):
    # type: (Namespace) -> DependencyConfiguration
    """Creates a dependency configuration from options registered by `register`."""

    return DependencyConfiguration.create(excluded=options.excluded, overridden=options.overridden)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path
import subprocess
import sys

from pex.compatibility import commonpath
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING
from pex.variables import ENV
from testing import PY310, ensure_python_interpreter, run_pex_command

if TYPE_CHECKING:
    from typing import Any


def test_boot_identification_leak(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    def assert_no_isolated_leak(python):
        # type: (str) -> None
        with ENV.patch(PEX_ROOT=pex_root), PythonInterpreter._cleared_memory_cache():
            interpreter = PythonInterpreter.from_binary(python)
            assert not any(
                pex_root == commonpath((pex_root, entry)) for entry in interpreter.sys_path
            ), (
                "The cached interpreter info for {python} contains leaked entries:\n"
                "{entries}".format(python=python, entries="\n".join(interpreter.sys_path))
            )

    empty_pex = os.path.join(str(tmpdir), "empty.pex")
    run_pex_command(
        args=["--pex-root", pex_root, "--runtime-pex-root", pex_root, "-o", empty_pex],
        python=sys.executable,
    ).assert_success()
    assert_no_isolated_leak(sys.executable)

    subprocess.check_call(args=[sys.executable, empty_pex, "-c", ""])
    assert_no_isolated_leak(sys.executable)

    other_python = ensure_python_interpreter(PY310)
    subprocess.check_call(args=[other_python, empty_pex, "-c", ""])
    # N.B.: Prior to the fix, this test failed with a vendored attrs leak:
    # E           AssertionError: The cached interpreter info for /home/jsirois/.pex_dev/pyenv/versions/3.10.7/bin/python3.10 contains leaked entries:
    # E             /tmp/pytest-of-jsirois/pytest-10/test_boot_identification_leak0/pex_root/isolated/975c556eea71292a09d930db2ca41875066d8be6/pex/vendor/_vendored/attrs
    # E             /home/jsirois/.pex_dev/pyenv/versions/3.10.7/lib/python310.zip
    # E             /home/jsirois/.pex_dev/pyenv/versions/3.10.7/lib/python3.10
    # E             /home/jsirois/.pex_dev/pyenv/versions/3.10.7/lib/python3.10/lib-dynload
    # E             /home/jsirois/.pex_dev/pyenv/versions/3.10.7/lib/python3.10/site-packages
    # E           assert not True
    # E            +  where True = any(<generator object test_boot_identification_leak.<locals>.assert_no_isolated_leak.<locals>.<genexpr> at 0x7fa5d084bc40>)
    assert_no_isolated_leak(other_python)



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from textwrap import dedent

import pytest

from pex import targets
from pex.interpreter import PythonInterpreter
from pex.pip.version import PipVersion, PipVersionValue
from pex.typing import TYPE_CHECKING
from testing import IntegResults, run_pex_command

if TYPE_CHECKING:
    from typing import NoReturn, Union


def pex_execute_cowsay(*extra_pex_args):
    # type: (*str) -> IntegResults
    return run_pex_command(
        args=list(extra_pex_args) + ["cowsay==5.0", "-c", "cowsay", "--", "Moo!"], quiet=True
    )


WARNING_PREFIX = "PEXWarning: "


def test_default_resolve_no_warning():
    # type: () -> None

    result = pex_execute_cowsay()
    result.assert_success()
    assert WARNING_PREFIX not in result.error
    assert "Moo!" in result.output


@pytest.fixture
def incompatible_pip_version():
    # type: () -> Union[PipVersionValue, NoReturn]
    for pip_version in PipVersion.values():
        if not pip_version.requires_python_applies(targets.current()):
            return pip_version

    pytest.skip(
        "There is no supported Pip version incompatible with {interpreter}.".format(
            interpreter=PythonInterpreter.get()
        )
    )
    raise AssertionError("Unreachable: satisfy type checker.")


def expected_incompatible_pip_message(
    incompatible_pip_version,  # type: PipVersionValue
    warning,  # type: bool
):
    # type: (...) -> str
    header = (
        "The Pip requested for PEX building was {pip_requirement} but it does not work with any "
        "of the targets selected.".format(pip_requirement=incompatible_pip_version.requirement)
    )
    return dedent(
        """\
        {prefix}{header}

        Pip {pip_version} requires Python {python_req} and the following target does not apply:
        1. {target}
        """.format(
            prefix=WARNING_PREFIX if warning else "",
            header=header,
            pip_version=incompatible_pip_version,
            python_req=incompatible_pip_version.requires_python,
            target=targets.current(),
        )
    )


def test_incompatible_resolve_warning(incompatible_pip_version):
    # type: (PipVersionValue) -> None

    result = pex_execute_cowsay("--pip-version", str(incompatible_pip_version))
    result.assert_success()
    assert (
        expected_incompatible_pip_message(incompatible_pip_version, warning=True) in result.error
    ), result.error
    assert "Moo!" in result.output, result.output


def test_incompatible_resolve_error(incompatible_pip_version):
    # type: (PipVersionValue) -> None

    result = pex_execute_cowsay(
        "--pip-version", str(incompatible_pip_version), "--no-allow-pip-version-fallback"
    )
    result.assert_failure()
    assert result.error.endswith(
        expected_incompatible_pip_message(incompatible_pip_version, warning=False)
    ), result.error



#!/usr/bin/env python3

from __future__ import annotations

import logging
import subprocess
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path, PurePath

import coloredlogs
import yaml


def build_cache_image(
    tox_envs: list[str], tag: str, pex_repo: str, git_ref: str, push: bool = False
) -> None:
    image_tag = f"ghcr.io/pex-tool/pex/cache:{tag}"
    subprocess.run(
        args=[
            "docker",
            "build",
            "--progress",
            "plain",
            "--build-arg",
            f"PEX_REPO={pex_repo}",
            "--build-arg",
            f"GIT_REF={git_ref}",
            "--build-arg",
            f"TOX_ENVS={','.join(tox_envs)}",
            "--tag",
            image_tag,
            str(PurePath("docker") / "cache"),
        ],
        check=True,
    )
    if push:
        subprocess.run(args=["docker", "push", image_tag], check=True)


def main() -> None:
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=(
            "Builds (and optionally pushes) a data-only cache image for use with "
            "`CACHE_MODE=pull ./dtox.sh ...`."
        ),
    )
    parser.add_argument(
        "-l",
        "--log-level",
        type=lambda arg: arg.upper(),
        default="INFO",
        choices=["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level (case insensitive).",
    )
    parser.add_argument("--color", default=None, action="store_true", help="Force colored logging.")
    parser.add_argument(
        "--tag",
        type=str,
        default="latest",
        help="The tag for the ghcr.io/pex-tool/pex/cache-all image.",
    )
    parser.add_argument(
        "--pex-repo",
        type=str,
        default="https://github.com/pex-tool/pex",
        help="The pex repo to clone and use for the docker/cache population.",
    )
    parser.add_argument(
        "--git-ref",
        type=str,
        default="HEAD",
        help="The git ref to use within `--pex-repo`.",
    )
    parser.add_argument(
        "--push",
        default=False,
        action="store_true",
        help="Push the image to the registry after building and tagging it.",
    )

    options = parser.parse_args()

    coloredlogs.install(
        level=options.log_level, fmt="%(levelname)s %(message)s", isatty=options.color
    )
    logger = logging.getLogger(parser.prog)
    logger.log(
        logging.root.level, "Logging configured at level {level}.".format(level=options.log_level)
    )

    with (Path(".github") / "workflows" / "ci.yml").open() as fp:
        data = yaml.full_load(fp)
    tox_envs = data["jobs"]["linux-tests"]["strategy"]["matrix"]["tox-env"]

    logger.info(f"Building caches for {len(tox_envs)} tox environments.")
    for tox_env in tox_envs:
        logger.debug(tox_env)

    build_cache_image(
        tox_envs,
        tag=options.tag,
        pex_repo=options.pex_repo,
        git_ref=options.git_ref,
        push=options.push,
    )


if __name__ == "__main__":
    main()



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import Any, Dict, List, Optional

from sphinx.application import Sphinx
from sphinx_pex.vars import Vars

PROJECT_ROOT = Path(__file__).parent.parent.parent.parent
ICON_ASSETS = PROJECT_ROOT / "assets"
DOC_ROOT = PROJECT_ROOT / "docs"
STATIC_ASSETS = [DOC_ROOT / "_static", DOC_ROOT / "_static_dynamic"]


def html_static_path() -> List[str]:
    return [
        static_asset_root.relative_to(DOC_ROOT).as_posix() for static_asset_root in STATIC_ASSETS
    ]


@dataclass(frozen=True)
class SVGIcon:
    @classmethod
    def load(cls, name: str, icon_asset: PurePath, url: str, css_class: str = "") -> SVGIcon:
        return cls(
            name=name, url=url, html=(ICON_ASSETS / icon_asset).read_text(), css_class=css_class
        )

    @classmethod
    def load_if_static_asset_exists(
        cls, name: str, icon_asset: PurePath, static_asset: PurePath, css_class: str = ""
    ) -> Optional[SVGIcon]:
        for static_asset_root in STATIC_ASSETS:
            static_asset_file = static_asset_root / static_asset
            if static_asset_file.is_file():
                return cls.load(
                    name=name,
                    icon_asset=icon_asset,
                    # N.B.: No matter where in the `html_static_path` the static asset comes from, its destination in
                    # the generated doc site will be the `_static/` dir.
                    url=(PurePath("_static") / static_asset).as_posix(),
                    css_class=css_class,
                )
        return None

    name: str
    url: str
    html: str
    css_class: str = ""

    def as_furo_footer_icon(self) -> Dict[str, str]:
        return {
            "name": self.name,
            "url": self.url,
            "html": self.html,
            "class": self.css_class,
        }


def setup(app: Sphinx) -> Dict[str, Any]:
    app.add_directive("vars", Vars)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path

from pex.common import CopyMode
from pex.typing import TYPE_CHECKING
from testing.cli import run_pex3
from testing.venv import assert_venv_site_packages_copy_mode

if TYPE_CHECKING:
    from typing import Any


def test_venv_site_packages_copies(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    expected_files = [
        os.path.join("cowsay", "__init__.py"),
        os.path.join("cowsay", "__main__.py"),
        os.path.join("cowsay", "characters.py"),
        os.path.join("cowsay", "main.py"),
        os.path.join("cowsay", "test.py"),
    ]

    venv = os.path.join(str(tmpdir), "venv")
    run_pex3("venv", "create", "--pex-root", pex_root, "-d", venv, "cowsay==5.0")
    assert_venv_site_packages_copy_mode(
        venv, expected_copy_mode=CopyMode.LINK, expected_files=expected_files
    )

    venv_copies = os.path.join(str(tmpdir), "venv-copies")
    run_pex3(
        "venv",
        "create",
        "--pex-root",
        pex_root,
        "-d",
        venv_copies,
        "cowsay==5.0",
        "--site-packages-copies",
    )
    assert_venv_site_packages_copy_mode(
        venv_copies, expected_copy_mode=CopyMode.COPY, expected_files=expected_files
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import re
from email.message import Message

from pex.dist_metadata import (
    DistMetadata,
    MetadataFiles,
    MetadataType,
    load_metadata,
    parse_message,
)
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Dict, Text

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class WheelMetadataLoadError(Exception):
    """Indicates an error loading WHEEL metadata."""


@attr.s(frozen=True)
class WHEEL(object):
    """The .whl WHEEL metadata file.

    See item 6 here for the WHEEL file contents: https://peps.python.org/pep-0427/#file-contents
    """

    _CACHE = {}  # type: Dict[Text, WHEEL]

    @classmethod
    def load(cls, location):
        # type: (Text) -> WHEEL
        wheel = cls._CACHE.get(location)
        if not wheel:
            metadata_files = load_metadata(location, restrict_types_to=(MetadataType.DIST_INFO,))
            if not metadata_files:
                raise WheelMetadataLoadError(
                    "Could not find any metadata in {wheel}.".format(wheel=location)
                )

            metadata_bytes = metadata_files.read("WHEEL")
            if not metadata_bytes:
                raise WheelMetadataLoadError(
                    "Could not find WHEEL metadata in {wheel}.".format(wheel=location)
                )
            metadata = parse_message(metadata_bytes)
            wheel = cls(files=metadata_files, metadata=metadata)
            cls._CACHE[location] = wheel
        return wheel

    files = attr.ib()  # type: MetadataFiles
    metadata = attr.ib()  # type: Message

    @property
    def root_is_purelib(self):
        # type: () -> bool

        # See:
        #   https://peps.python.org/pep-0427/#installing-a-wheel-distribution-1-0-py32-none-any-whl
        return cast(bool, "true" == self.metadata.get("Root-Is-Purelib"))


@attr.s(frozen=True)
class Wheel(object):
    @classmethod
    def load(cls, wheel_path):
        # type: (str) -> Wheel

        metadata = WHEEL.load(wheel_path)

        metadata_path = metadata.files.metadata_file_rel_path("WHEEL")
        if not metadata_path:
            raise WheelMetadataLoadError(
                "Could not find WHEEL metadata in {wheel}.".format(wheel=wheel_path)
            )

        wheel_metadata_dir = os.path.dirname(metadata_path)
        if not wheel_metadata_dir.endswith(".dist-info"):
            raise WheelMetadataLoadError(
                "Expected WHEEL metadata for {wheel} to be housed in a .dist-info directory, but was "
                "found at {wheel_metadata_path}.".format(
                    wheel=wheel_path, wheel_metadata_path=metadata_path
                )
            )
        # Although not crisply defined, all PEPs lead to PEP-508 which restricts project names
        # to ASCII: https://peps.python.org/pep-0508/#names. Likewise, version numbers are also
        # restricted to ASCII: https://peps.python.org/pep-0440/. Since the `.dist-info` dir
        # path is defined as `<project name>-<version>.dist-info` in
        # https://peps.python.org/pep-0427/, we are safe in assuming ASCII overall for the wheel
        # metadata dir path.
        metadata_dir = str(wheel_metadata_dir)

        data_dir = re.sub(r"\.dist-info$", ".data", metadata_dir)

        return cls(
            location=wheel_path,
            metadata_dir=metadata_dir,
            metadata_files=metadata.files,
            metadata=metadata,
            data_dir=data_dir,
        )

    location = attr.ib()  # type: str
    metadata_dir = attr.ib()  # type: str
    metadata_files = attr.ib()  # type: MetadataFiles
    metadata = attr.ib()  # type: WHEEL
    data_dir = attr.ib()  # type: str

    @property
    def root_is_purelib(self):
        # type: () -> bool
        return self.metadata.root_is_purelib

    def dist_metadata(self):
        return DistMetadata.from_metadata_files(self.metadata_files)

    def metadata_path(self, *components):
        # typ: (*str) -> str
        return os.path.join(self.metadata_dir, *components)

    def data_path(self, *components):
        # typ: (*str) -> str
        return os.path.join(self.data_dir, *components)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import pytest

from pex.typing import TYPE_CHECKING
from testing.docker import DockerVirtualenvRunner

if TYPE_CHECKING:
    from typing import Any


@pytest.fixture
def fedora39_virtualenv_runner(tmpdir):
    # type: (Any) -> DockerVirtualenvRunner

    return DockerVirtualenvRunner.create(
        base_image="fedora:39",
        python="python3.12",
        virtualenv_version="20.25.1",
        tmpdir=str(tmpdir),
    )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import shutil
import subprocess
import tempfile
from textwrap import dedent

import colors
import pytest

from pex.resolve.lockfile import json_codec
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any, Callable, List


@pytest.mark.parametrize(
    ["archive_pex_requirement"],
    [
        pytest.param(
            "https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip#egg=cowsay",
            id="Create Pex [Pip Proprietary]",
        ),
        pytest.param(
            "cowsay @ https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip",
            id="Create Pex [PEP-508]",
        ),
    ],
)
def test_pex_archive_direct_reference(
    tmpdir,  # type: Any
    archive_pex_requirement,  # type: str
):
    # type: (...) -> None

    result = run_pex_command(
        args=[
            archive_pex_requirement,
            "-c",
            "cowsay",
            "--",
            "Moo!",
        ]
    )
    result.assert_success()
    assert "Moo!" in result.output


@pytest.mark.parametrize(
    ["archive_lock_requirement"],
    [
        pytest.param(
            "https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip#egg=cowsay",
            id="Create Lock [Pip Proprietary]",
        ),
        pytest.param(
            "cowsay @ https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip",
            id="Create Lock [PEP-508]",
        ),
    ],
)
@pytest.mark.parametrize(
    ["archive_pex_requirements"],
    [
        pytest.param(
            ["https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip#egg=cowsay"],
            id="Subset [Pip Proprietary]",
        ),
        pytest.param(
            ["cowsay @ https://github.com/VaasuDevanS/cowsay-python/archive/v5.0.zip"],
            id="Subset [PEP-508]",
        ),
        pytest.param([], id="Full"),
    ],
)
def test_lock_create_archive_direct_reference(
    tmpdir,  # type: Any
    archive_lock_requirement,  # type: str
    archive_pex_requirements,  # type: List[str]
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        archive_lock_requirement,
        "--indent",
        "2",
        "-o",
        lock,
    ).assert_success()

    def assert_create_and_run_pex_from_lock():
        # type: () -> None
        result = run_pex_command(
            args=archive_pex_requirements
            + [
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--lock",
                lock,
                "-c",
                "cowsay",
                "--",
                "Moo!",
            ]
        )
        result.assert_success()
        assert "Moo!" in result.output

    assert_create_and_run_pex_from_lock()
    shutil.rmtree(pex_root)
    assert_create_and_run_pex_from_lock()


@pytest.mark.parametrize(
    ["create_local_project_lock_requirement"],
    [
        pytest.param(lambda clone_dir: clone_dir, id="Create Lock [Pip Proprietary]"),
        pytest.param(
            lambda clone_dir: "ansicolors @ file://{}".format(clone_dir), id="Create Lock [PEP-508]"
        ),
    ],
)
@pytest.mark.parametrize(
    ["create_local_project_pex_requirements"],
    [
        pytest.param(lambda clone_dir: [clone_dir], id="Subset [Pip Proprietary]"),
        pytest.param(
            lambda clone_dir: ["ansicolors @ file://{}".format(clone_dir)], id="Subset [PEP-508]"
        ),
        pytest.param(lambda clone_dir: [], id="Full"),
    ],
)
def test_lock_create_local_project_direct_reference(
    tmpdir,  # type: Any
    create_local_project_lock_requirement,  # type: Callable[[str], str]
    create_local_project_pex_requirements,  # type: Callable[[str], List[str]]
):
    # type: (...) -> None

    clone_dir = os.path.join(str(tmpdir), "ansicolors")
    subprocess.check_call(args=["git", "init", clone_dir])

    ansicolors_1_1_8_sha = "c965f5b9103c5bd32a1572adb8024ebe83278fb0"
    subprocess.check_call(
        args=[
            "git",
            "fetch",
            "--depth",
            "1",
            "https://github.com/jonathaneunice/colors",
            ansicolors_1_1_8_sha,
        ],
        cwd=clone_dir,
    )
    subprocess.check_call(args=["git", "reset", "--hard", ansicolors_1_1_8_sha], cwd=clone_dir)

    lock_requirement = create_local_project_lock_requirement(clone_dir)
    pex_requirements = create_local_project_pex_requirements(clone_dir)

    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        lock_requirement,
        "--indent",
        "2",
        "-o",
        lock,
    ).assert_success()

    def assert_create_and_run_pex_from_lock():
        # type: () -> None
        result = run_pex_command(
            args=pex_requirements
            + [
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--lock",
                lock,
                "--",
                "-c",
                "import colors; print(colors.yellow('Vogon Constructor Fleet!'))",
            ]
        )
        result.assert_success()
        assert colors.yellow("Vogon Constructor Fleet!") == result.output.strip()

    assert_create_and_run_pex_from_lock()
    shutil.rmtree(pex_root)
    assert_create_and_run_pex_from_lock()

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(
            dedent(
                """\
                diff --git a/setup.py b/setup.py
                index 0b58889..bdb7c90 100755
                --- a/setup.py
                +++ b/setup.py
                @@ -42,3 +42,4 @@ setup(
                         'Topic :: Software Development :: Libraries :: Python Modules'
                     ]
                 )
                +# Changed
                """
            ).encode("utf-8")
        )
        fp.flush()
        subprocess.check_call(args=["git", "apply", fp.name], cwd=clone_dir)

    # We patched the source but have a cached wheel built from it before the patch in
    # ~/.pex/installed_wheels; so no "download" is performed.
    assert_create_and_run_pex_from_lock()

    # But now we do need to "download" the project, build a wheel and install it. The hash check
    # should fail.
    shutil.rmtree(pex_root)
    result = run_pex_command(
        args=["--pex-root", pex_root, "--runtime-pex-root", pex_root, "--lock", lock]
    )
    result.assert_failure()

    lockfile = json_codec.load(lockfile_path=lock)
    assert 1 == len(lockfile.locked_resolves)
    locked_resolve = lockfile.locked_resolves[0]
    assert 1 == len(locked_resolve.locked_requirements)
    locked_requirement = locked_resolve.locked_requirements[0]
    assert (
        dedent(
            """\
            There was 1 error downloading required artifacts:
            1. ansicolors 1.1.8 from file://{clone_dir}
                Expected sha256 hash of {expected} when downloading ansicolors but hashed to
            """
        )
        .format(
            clone_dir=clone_dir,
            expected=locked_requirement.artifact.fingerprint.hash,
        )
        .strip()
        in result.error
    ), result.error



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import hatchling.build
import pex_build

# We re-export all hatchling's PEP-517 build backend hooks here for the build frontend to call.
from hatchling.build import *  # NOQA

if pex_build.TYPE_CHECKING:
    from typing import Any, Dict, List, Optional


def get_requires_for_build_wheel(config_settings=None):
    # type: (Optional[Dict[str, Any]]) -> List[str]

    reqs = hatchling.build.get_requires_for_build_wheel(
        config_settings=config_settings
    )  # type: List[str]
    if pex_build.INCLUDE_DOCS:
        with open("docs-requirements.txt") as fp:
            for raw_req in fp.readlines():
                req = raw_req.strip()
                if not req or req.startswith("#"):
                    continue
                reqs.append(req)
    return reqs



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import re

from pex.third_party.packaging.utils import canonicalize_name
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Text

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def _ensure_ascii_str(text):
    # type: (Text) -> str

    # Although not crisply defined, all PEPs lead to PEP-508 which restricts project names to
    # ASCII: https://peps.python.org/pep-0508/#names
    return str(text)


@attr.s(frozen=True)
class ProjectName(object):
    """Encodes a canonicalized project name as per PEP-503.

    See: https://www.python.org/dev/peps/pep-0503/#normalized-names
    """

    class InvalidError(ValueError):
        """Indicates an invalid project name as per https://peps.python.org/pep-0508/#names."""

    # See: https://peps.python.org/pep-0508/#names
    _VALID_RE = re.compile(r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", re.IGNORECASE)

    raw = attr.ib(eq=False, converter=_ensure_ascii_str)  # type: str
    validated = attr.ib(eq=False, default=False)  # type: bool
    normalized = attr.ib(init=False)  # type: str

    def __attrs_post_init__(self):
        if self.validated and not self._VALID_RE.match(self.raw):
            raise self.InvalidError(
                "The given project name {value!r} is not a valid. It must conform to the regex "
                "{pattern!r} as specified in https://peps.python.org/pep-0508/#names".format(
                    value=self.raw, pattern=self._VALID_RE.pattern
                )
            )
        object.__setattr__(self, "normalized", canonicalize_name(self.raw))

    def __str__(self):
        # type: () -> str
        return self.normalized



#!/usr/bin/env python
# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import atexit
import logging
import os
import re
import subprocess
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import coloredlogs

# Ensure the repo root is on the `sys.path` (for access to the pex and testing packages).
os.environ["_PEX_TEST_PROJECT_DIR"] = str(
    subprocess.check_output(["git", "rev-parse", "--show-toplevel"]).decode("ascii").strip()
)
sys.path.insert(0, os.environ["_PEX_TEST_PROJECT_DIR"])

from pex.compatibility import urlparse
from pex.typing import TYPE_CHECKING, cast
from testing import devpi, pex_project_dir

if TYPE_CHECKING:
    from typing import Iterator, Tuple


def iter_test_control_env_vars():
    # type: () -> Iterator[Tuple[str, str]]
    for var, value in sorted(os.environ.items()):
        if re.search(r"(PEX|PIP|PYTHON)", var) and var != "PYTHONHASHSEED":
            yield var, value


def main():
    # type: () -> int
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-l",
        "--log-level",
        type=lambda arg: arg.upper(),
        default="INFO",
        choices=["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level (case insensitive).",
    )
    parser.add_argument("--color", default=None, action="store_true", help="Force colored logging.")
    parser.add_argument(
        "--devpi", action="store_true", help="Proxy PyPI through a local devpi server."
    )
    parser.add_argument(
        "--require-devpi",
        action="store_true",
        help=(
            "Fail fast if `--devpi` was requested but the server can't be started or connected to. "
            "Exits with code 42."
        ),
    )
    parser.add_argument(
        "--devpi-host",
        type=str,
        default="0.0.0.0",
        help="The domain/ip address to have the local devpi server listen on.",
    )
    parser.add_argument(
        "--devpi-port", type=int, default=0, help="The port to bind the local devpi server to."
    )
    parser.add_argument(
        "--devpi-timeout",
        type=float,
        default=5.0,
        help=(
            "The maximum amount of time (in seconds) to wait for devpi server to start before "
            "continuing without it."
        ),
    )
    parser.add_argument(
        "--devpi-max-connection-retries",
        type=int,
        default=3,
        help="The maximum number of PyPI connection attempt retries.",
    )
    parser.add_argument(
        "--devpi-request-timeout",
        type=int,
        default=5,
        help=(
            "The maximum amount of time to wait (in seconds) for request activity before "
            "terminating the request."
        ),
    )
    parser.add_argument(
        "--shutdown-devpi",
        action="store_true",
        help="If using a devpi server for the run, shut it down at the end of the run.",
    )
    parser.add_argument("--it", action="store_true", help="Restrict scope to integration tests.")
    options, passthrough_args = parser.parse_known_args()

    coloredlogs.install(
        level=options.log_level, fmt="%(levelname)s %(message)s", isatty=options.color
    )
    logger = logging.getLogger(parser.prog)
    logger.log(
        logging.root.level, "Logging configured at level {level}.".format(level=options.log_level)
    )

    if options.devpi:
        if options.shutdown_devpi:
            atexit.register(devpi.shutdown)
        launch_result = devpi.launch(
            host=options.devpi_host,
            port=options.devpi_port,
            timeout=options.devpi_timeout,
            max_connection_retries=options.devpi_max_connection_retries,
            request_timeout=options.devpi_request_timeout,
        )
        if isinstance(launch_result, devpi.LaunchResult):
            os.environ["_PEX_USE_PIP_CONFIG"] = str(True)
            os.environ["PIP_INDEX_URL"] = launch_result.url
            os.environ["PIP_TRUSTED_HOST"] = cast(
                # We know the local devpi server URL will always have a host and never be None.
                str,
                urlparse.urlparse(launch_result.url).hostname,
            )
            logger.info(
                "Devpi server already running."
                if launch_result.already_running
                else "Launched devpi server."
            )
        else:
            if options.require_devpi:
                logger.critical("Failed to launch devpi server.")
                log_log_level = logging.ERROR
            else:
                logger.warning("Failed to launch devpi server. Continuing without it...")
                log_log_level = logging.DEBUG
            with open(launch_result) as fp:
                for line in fp:
                    logger.log(log_log_level, line.rstrip())
            if options.require_devpi:
                return 42

    test_control_env_vars = list(iter_test_control_env_vars())
    if test_control_env_vars:
        logger.info("Test control environment variables:")
        for var, value in test_control_env_vars:
            logger.info("{var}={value}".format(var=var, value=value))
    else:
        logger.info("No test control environment variables set.")

    if options.it:
        pytest_args = ["-n", "auto", "tests/integration"]
    else:
        pytest_args = ["tests", "--ignore", "tests/integration"]

    return subprocess.call(
        args=[sys.executable, "-m", "pytest"] + pytest_args + passthrough_args or ["-vvs"],
        cwd=pex_project_dir(),
    )


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import itertools
import os.path

from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve.locked_resolve import FileArtifact
from pex.resolve.lockfile import json_codec
from pex.resolve.resolved_requirement import Pin
from pex.typing import TYPE_CHECKING
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


def test_update_sdists_not_updated(tmpdir):
    # type: (Any) -> None

    constraints = os.path.join(str(tmpdir), "constraints.txt")
    with open(constraints, "w") as fp:
        print("ansicolors<1.1.8", file=fp)
        print("cowsay<6", file=fp)

    lock = os.path.join(str(tmpdir), "lock.json")

    def assert_lock(*pins):
        # type: (*Pin) -> None

        lockfile = json_codec.load(lock)
        assert 1 == len(lockfile.locked_resolves)
        locked_resolve = lockfile.locked_resolves[0]
        locked_requirements = {
            locked_req.pin: tuple(locked_req.iter_artifacts())
            for locked_req in locked_resolve.locked_requirements
        }
        assert set(pins) == set(locked_requirements)
        assert all(
            isinstance(artifact, FileArtifact) and artifact.is_source
            for artifact in itertools.chain.from_iterable(locked_requirements.values())
        )

    run_pex3(
        "lock",
        "create",
        "--no-wheel",
        "--constraints",
        constraints,
        "ansicolors",
        "cowsay",
        "--indent",
        "2",
        "-o",
        lock,
    ).assert_success()
    assert_lock(
        Pin(ProjectName("ansicolors"), Version("1.1.7")), Pin(ProjectName("cowsay"), Version("5.0"))
    )

    # N.B.: Pre-fix this test would lead to an artifact comparison assertion for cowsay, which is
    # expected to be unmodified by the lock update.
    #
    # E       Traceback (most recent call last):
    # E         File "/home/jsirois/dev/pex-tool/pex/pex/result.py", line 105, in catch
    # E           return func(*args, **kwargs)
    # E                  ^^^^^^^^^^^^^^^^^^^^^
    # E         File "/home/jsirois/dev/pex-tool/pex/pex/resolve/lockfile/updater.py", line 320, in update_resolve
    # E           assert updated_requirement.artifact == locked_requirement.artifact
    # E                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # E       AssertionError
    # E       Encountered 1 error updating /tmp/pytest-of-jsirois/pytest-8/test_update_sdists_not_updated0/lock.json:
    # E       1.) cp311-cp311-manylinux_2_35_x86_64:
    run_pex3("lock", "update", "-v", "-p", "ansicolors<1.1.9", lock).assert_success()
    assert_lock(
        Pin(ProjectName("ansicolors"), Version("1.1.8")), Pin(ProjectName("cowsay"), Version("5.0"))
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

# When running under MyPy, this will be set to True for us automatically; so we can use it as a typing module import
# guard to protect Python 2 imports of typing - which is not normally available in Python 2.
TYPE_CHECKING = False


INCLUDE_DOCS = os.environ.get("__PEX_BUILD_INCLUDE_DOCS__", "False").lower() in ("1", "true")



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex import resolver
from pex.dist_metadata import Requirement
from pex.pep_427 import InstallableType
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve import lock_resolver
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolver_configuration import PipConfiguration, ReposConfiguration, ResolverVersion
from pex.resolve.resolvers import Resolver, ResolveResult
from pex.result import try_
from pex.targets import Targets
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Optional, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


_DEFAULT_REPOS = ReposConfiguration.create()


@attr.s(frozen=True)
class ConfiguredResolver(Resolver):
    @classmethod
    def version(cls, pip_version):
        # type: (PipVersionValue) -> ConfiguredResolver
        return cls(
            PipConfiguration(
                version=pip_version, resolver_version=ResolverVersion.default(pip_version)
            )
        )

    @classmethod
    def default(cls):
        # type: () -> ConfiguredResolver
        return cls.version(PipVersion.DEFAULT)

    pip_configuration = attr.ib()  # type: PipConfiguration

    def is_default_repos(self):
        return self.pip_configuration.repos_configuration == _DEFAULT_REPOS

    def resolve_lock(
        self,
        lock,  # type: Lockfile
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
        result_type=InstallableType.INSTALLED_WHEEL_CHROOT,  # type: InstallableType.Value
    ):
        # type: (...) -> ResolveResult
        return try_(
            lock_resolver.resolve_from_lock(
                targets=targets,
                lock=lock,
                resolver=self,
                indexes=self.pip_configuration.repos_configuration.indexes,
                find_links=self.pip_configuration.repos_configuration.find_links,
                resolver_version=self.pip_configuration.resolver_version,
                network_configuration=self.pip_configuration.network_configuration,
                build_configuration=self.pip_configuration.build_configuration,
                compile=False,
                transitive=self.pip_configuration.transitive,
                verify_wheels=True,
                max_parallel_jobs=self.pip_configuration.max_jobs,
                pip_version=pip_version or self.pip_configuration.version,
                use_pip_config=self.pip_configuration.use_pip_config,
                extra_pip_requirements=self.pip_configuration.extra_requirements,
                result_type=result_type,
            )
        )

    def resolve_requirements(
        self,
        requirements,  # type: Iterable[str]
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
        transitive=None,  # type: Optional[bool]
        extra_resolver_requirements=None,  # type: Optional[Tuple[Requirement, ...]]
        result_type=InstallableType.INSTALLED_WHEEL_CHROOT,  # type: InstallableType.Value
    ):
        # type: (...) -> ResolveResult
        return resolver.resolve(
            targets=targets,
            requirements=requirements,
            allow_prereleases=False,
            transitive=transitive if transitive is not None else self.pip_configuration.transitive,
            indexes=self.pip_configuration.repos_configuration.indexes,
            find_links=self.pip_configuration.repos_configuration.find_links,
            resolver_version=self.pip_configuration.resolver_version,
            network_configuration=self.pip_configuration.network_configuration,
            build_configuration=self.pip_configuration.build_configuration,
            compile=False,
            max_parallel_jobs=self.pip_configuration.max_jobs,
            ignore_errors=False,
            verify_wheels=True,
            pip_version=pip_version or self.pip_configuration.version,
            resolver=self,
            use_pip_config=self.pip_configuration.use_pip_config,
            extra_pip_requirements=extra_resolver_requirements
            if extra_resolver_requirements is not None
            else self.pip_configuration.extra_requirements,
            result_type=result_type,
        )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path
import sys

import pytest

from pex.compatibility import commonpath
from pex.resolve.resolver_configuration import ResolverVersion
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 5),
    reason=(
        "This test relies on the python-forge 18.6.0 offering only the single "
        "python_forge-18.6.0-py35-none-any.whl distribution on PyPI and this wheel only works for "
        "Python 3.5 and greater"
    ),
)
@pytest.mark.parametrize(
    "resolver_version",
    [
        pytest.param(resolver_version, id=str(resolver_version))
        for resolver_version in ResolverVersion.values()
        if ResolverVersion.applies(resolver_version)
    ],
)
def test_abi_none_locking(
    tmpdir,  # type: Any
    resolver_version,  # type: ResolverVersion.Value
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--style",
        "universal",
        "--resolver-version",
        str(resolver_version),
        "--interpreter-constraint",
        "==3.11.*",
        "python-forge==18.6.0",
        "-o",
        lock,
        "--indent",
        "2",
        "--pex-root",
        pex_root,
    ).assert_success()

    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            lock,
            "--",
            "-c",
            "import forge; print(forge.__file__)",
        ]
    )
    result.assert_success()
    assert pex_root == commonpath([pex_root, result.output.strip()])



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import itertools
import os.path
import sys
from argparse import Action, ArgumentError, ArgumentParser, ArgumentTypeError, _ActionsContainer
from collections import OrderedDict
from operator import attrgetter

from pex import dependency_configuration, pex_warnings
from pex.argparse import HandleBoolAction
from pex.cli.command import BuildTimeCommand
from pex.commands.command import JsonMixin, OutputMixin
from pex.common import is_exe, pluralize, safe_delete, safe_open
from pex.compatibility import commonpath, shlex_quote
from pex.dependency_configuration import DependencyConfiguration
from pex.dist_metadata import (
    Constraint,
    Distribution,
    MetadataType,
    Requirement,
    RequirementParseError,
)
from pex.enum import Enum
from pex.exceptions import production_assert
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.pep_376 import InstalledWheel, Record
from pex.pep_427 import InstallableType
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve import project, requirement_options, resolver_options, target_options
from pex.resolve.config import finalize as finalize_resolve_config
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.lock_resolver import resolve_from_lock
from pex.resolve.locked_resolve import (
    LocalProjectArtifact,
    LockConfiguration,
    LockStyle,
    Resolved,
    TargetSystem,
    VCSArtifact,
)
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.create import create
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.lockfile.subset import subset
from pex.resolve.lockfile.updater import (
    ArtifactsUpdate,
    ArtifactUpdate,
    DeleteUpdate,
    FingerprintUpdate,
    LockUpdate,
    LockUpdater,
    ResolveUpdateRequest,
    VersionUpdate,
)
from pex.resolve.path_mappings import PathMappings
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolved_requirement import Fingerprint, Pin
from pex.resolve.resolver_configuration import LockRepositoryConfiguration, PipConfiguration
from pex.resolve.resolver_options import parse_lockfile
from pex.resolve.target_configuration import InterpreterConstraintsNotSatisfied, TargetConfiguration
from pex.result import Error, Ok, Result, try_
from pex.sorted_tuple import SortedTuple
from pex.targets import LocalInterpreter, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast
from pex.venv.virtualenv import InvalidVirtualenvError, Virtualenv
from pex.version import __version__

if TYPE_CHECKING:
    from typing import IO, Dict, Iterable, List, Mapping, Optional, Set, Text, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class FingerprintMismatch(Enum["FingerprintMismatch.Value"]):
    class Value(Enum.Value):
        pass

    IGNORE = Value("ignore")
    WARN = Value("warn")
    ERROR = Value("error")


class ExportFormat(Enum["ExportFormat.Value"]):
    class Value(Enum.Value):
        pass

    PIP = Value("pip")
    PIP_NO_HASHES = Value("pip-no-hashes")
    PEP_665 = Value("pep-665")


class ExportSortBy(Enum["ExportSortBy.Value"]):
    class Value(Enum.Value):
        pass

    SPECIFICITY = Value("specificity")
    PROJECT_NAME = Value("project-name")


class DryRunStyle(Enum["DryRunStyle.Value"]):
    class Value(Enum.Value):
        pass

    DISPLAY = Value("display")
    CHECK = Value("check")


class HandleDryRunAction(Action):
    def __init__(self, *args, **kwargs):
        kwargs["nargs"] = "?"
        super(HandleDryRunAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_str=None):
        if option_str.startswith("--no-"):
            if value:
                raise ArgumentError(
                    None,
                    "The {option} option does not take a value; given: {value!r}".format(
                        option=option_str, value=value
                    ),
                )
            dry_run_style = None
        elif value:
            try:
                dry_run_style = DryRunStyle.for_value(value)
            except ValueError:
                raise ArgumentTypeError(
                    "Invalid value for {option}: {value!r}. Either pass no value for {default!r} "
                    "or one of: {choices}".format(
                        option=option_str,
                        value=value,
                        default=DryRunStyle.DISPLAY,
                        choices=", ".join(map(repr, DryRunStyle.values())),
                    )
                )
        else:
            dry_run_style = DryRunStyle.DISPLAY
        setattr(namespace, self.dest, dry_run_style)


@attr.s(frozen=True)
class SyncTarget(object):
    _PIP_PROJECT_NAME = ProjectName("pip")

    @classmethod
    def resolve_command(
        cls,
        venv,  # type: Virtualenv
        command=None,  # type: Optional[Tuple[str, ...]]
    ):
        # type: (...) -> SyncTarget
        if command:
            argv0 = command[0] if os.path.isabs(command[0]) else venv.bin_path(command[0])
            command = (argv0,) + command[1:]
        return SyncTarget(venv=venv, command=command)

    @classmethod
    def resolve_venv(
        cls,
        argv0,  # type str
        *additional_args  # type: str
    ):
        # type: (...) -> Union[SyncTarget, Error]

        argv0_path = None  # type: Optional[str]
        if is_exe(argv0):
            argv0_path = argv0
        else:
            path = os.environ.get("PATH")
            if path:
                for entry in path.split(os.pathsep):
                    exe_path = os.path.abspath(os.path.join(entry, argv0))
                    if is_exe(exe_path):
                        argv0_path = exe_path
                        break

        if argv0_path:
            venv = Virtualenv.enclosing(python=argv0_path)
            if not venv:
                try:
                    venv = Virtualenv(os.path.dirname(os.path.dirname(argv0_path)))
                except InvalidVirtualenvError as e:
                    return Error(
                        "Could not find a valid venv enclosing {argv0} to sync: {err}.\n"
                        "Try explicitly specifying a venv to sync with `--venv`.".format(
                            err=e, argv0=argv0
                        )
                    )
            command = [argv0_path]
            command.extend(additional_args)
            return SyncTarget(venv=venv, command=tuple(command))

        if os.path.isdir(argv0) and not additional_args:
            try:
                return SyncTarget(venv=Virtualenv(argv0))
            except InvalidVirtualenvError as e:
                return Error(
                    "The directory at {path} is not a valid venv to sync: {err}.\n"
                    "Try explicitly specifying a venv to sync with `--venv`.".format(
                        err=e, path=argv0
                    )
                )

        return Error(
            "Could not determine a venv to sync after examining {argv0}.\n"
            "Try explicitly specifying a venv to sync with `--venv`.".format(argv0=argv0)
        )

    venv = attr.ib()  # type: Virtualenv
    command = attr.ib(default=None)  # type: Optional[Tuple[str, ...]]

    def sync(
        self,
        distributions,  # type: Iterable[Distribution]
        confirm=True,  # type: bool
        retain_pip=False,  # type: bool
    ):
        # type: (...) -> Result

        abs_venv_dir = os.path.realpath(self.venv.venv_dir)

        existing_distributions_by_project_name = {
            dist.metadata.project_name: dist for dist in self.venv.iter_distributions()
        }  # type: Dict[ProjectName, Distribution]
        installed_pip = existing_distributions_by_project_name.get(
            self._PIP_PROJECT_NAME
        )  # type: Optional[Distribution]

        resolved_pip = None  # type: Optional[Distribution]
        to_remove = []  # type: List[Distribution]
        to_install = []  # type: List[Distribution]
        for distribution in distributions:
            if self._PIP_PROJECT_NAME == distribution.metadata.project_name:
                resolved_pip = distribution
            existing_distribution = existing_distributions_by_project_name.pop(
                distribution.metadata.project_name, None
            )
            if not existing_distribution:
                to_install.append(distribution)
            elif existing_distribution.metadata.version != distribution.metadata.version:
                to_remove.append(existing_distribution)
                to_install.append(distribution)
        if retain_pip:
            existing_distributions_by_project_name.pop(self._PIP_PROJECT_NAME, None)
        to_remove.extend(existing_distributions_by_project_name.values())

        to_unlink_by_pin = (
            OrderedDict()
        )  # type: OrderedDict[Tuple[ProjectName, Version], List[Text]]
        for distribution in to_remove:
            to_unlink = []  # type: List[Text]
            if distribution.metadata.type is MetadataType.DIST_INFO:
                to_unlink.extend(
                    os.path.realpath(os.path.join(distribution.location, installed_file.path))
                    for installed_file in Record.read(distribution.iter_metadata_lines("RECORD"))
                )
            elif distribution.metadata.type is MetadataType.EGG_INFO:
                installed_files = distribution.metadata.files.metadata_file_rel_path(
                    "installed-files.txt"
                )
                if installed_files:
                    base_dir = os.path.realpath(
                        os.path.join(distribution.location, os.path.dirname(installed_files))
                    )
                    to_unlink.extend(
                        os.path.join(base_dir, path)
                        for path in distribution.iter_metadata_lines("installed-files.txt")
                    )
                    # Pip generates "installed-files.txt" upon installing and this file does not
                    # include itself; so we tack that on here ourselves.
                    to_unlink.append(
                        os.path.realpath(os.path.join(distribution.location, installed_files))
                    )
            if to_unlink:
                to_unlink_by_pin[
                    (distribution.metadata.project_name, distribution.metadata.version)
                ] = [file for file in to_unlink if abs_venv_dir == commonpath((abs_venv_dir, file))]
        if confirm and to_unlink_by_pin:
            for (project_name, version), files in to_unlink_by_pin.items():
                print(project_name, version, ":", file=sys.stderr)
                for f in files:
                    print("    ", f, file=sys.stderr)
            if input(
                "Remove the outdated distributions listed above from the venv at "
                "{venv}? [yN]: ".format(venv=self.venv.venv_dir)
            ).strip().lower() not in ("y", "yes"):
                return Error("Sync cancelled.")

        if to_unlink_by_pin:
            parent_dirs = set()  # type: Set[Text]
            for file in itertools.chain.from_iterable(to_unlink_by_pin.values()):
                safe_delete(file)
                parent_dirs.add(os.path.dirname(file))
            for parent_dir in sorted(parent_dirs, reverse=True):
                if not os.listdir(parent_dir) and abs_venv_dir == commonpath(
                    (abs_venv_dir, parent_dir)
                ):
                    os.rmdir(parent_dir)

        if retain_pip and not resolved_pip and not installed_pip:
            self.venv.ensure_pip(upgrade=True)

        if to_install:
            for distribution in to_install:
                for src, dst in InstalledWheel.load(distribution.location).reinstall_venv(
                    self.venv
                ):
                    TRACER.log("Installed {src} -> {dst}".format(src=src, dst=dst))
            for script in self.venv.rewrite_scripts():
                TRACER.log("Re-wrote script shebang for {script}".format(script=script))

        if self.command:
            try:
                os.execv(self.command[0], self.command)
            except OSError as e:
                return Error("Failed to execute {exe}: {err}".format(exe=self.command[0], err=e))

        return Ok()


@attr.s(frozen=True)
class LockUpdateRequest(object):
    targets = attr.ib()  # type: Targets
    _lock_file_path = attr.ib()  # type: str
    _lock_updater = attr.ib()  # type: LockUpdater
    _update_requests = attr.ib()  # type: Iterable[ResolveUpdateRequest]

    def update(
        self,
        updates=(),  # type: Iterable[Requirement]
        replacements=(),  # type: Iterable[Requirement]
        deletes=(),  # type: Iterable[ProjectName]
        artifacts_can_change=False,  # type: bool
    ):
        # type: (...) -> Union[LockUpdate, Result]
        if not self._update_requests:
            return self._no_updates()

        return self._lock_updater.update(
            update_requests=self._update_requests,
            updates=updates,
            replacements=replacements,
            deletes=deletes,
            artifacts_can_change=artifacts_can_change,
        )

    def sync(self, requirement_configuration):
        # type: (RequirementConfiguration) -> Union[LockUpdate, Result]
        if not self._update_requests:
            return self._no_updates()

        return self._lock_updater.sync(
            update_requests=self._update_requests,
            requirement_configuration=requirement_configuration,
        )

    def _no_updates(self):
        # type: () -> Ok
        return Ok(
            "No lock update was performed.\n"
            "The following platforms present in {lock_file} were not found on the local "
            "machine:\n"
            "{missing_platforms}\n"
            "You might still be able to update the lock by adjusting target options like "
            "--python-path.".format(
                lock_file=self._lock_file_path,
                missing_platforms="\n".join(
                    sorted(
                        "+ {platform}".format(platform=locked_resolve.platform_tag)
                        for locked_resolve in self._lock_updater.lock_file.locked_resolves
                    )
                ),
            )
        )


class Lock(OutputMixin, JsonMixin, BuildTimeCommand):
    """Operate on PEX lock files."""

    @staticmethod
    def _add_target_options(parser):
        # type: (_ActionsContainer) -> None
        target_options.register(
            parser.add_argument_group(
                title="Target options",
                description=(
                    "Specify which interpreters and platforms resolved distributions must support."
                ),
            )
        )

    @classmethod
    def _create_resolver_options_group(cls, parser):
        # type: (_ActionsContainer) -> _ActionsContainer
        return parser.add_argument_group(
            title="Resolver options",
            description="Configure how third party distributions are resolved.",
        )

    @classmethod
    def _add_resolve_options(cls, parser):
        # type: (_ActionsContainer) -> None
        options_group = parser.add_argument_group(
            title="Requirement options",
            description="Indicate which distributions should be resolved",
        )
        requirement_options.register(options_group)
        project.register_options(
            options_group,
            help=(
                "Add the transitive dependencies of the local project at the specified path to "
                "the lock but do not lock project itself."
            ),
        )
        dependency_configuration.register(options_group)
        cls._add_target_options(parser)
        resolver_options.register(
            cls._create_resolver_options_group(parser),
            include_pex_repository=False,
            include_lock=False,
        )

    @classmethod
    def _add_lockfile_option(
        cls,
        parser,  # type: _ActionsContainer
        verb,  # type: str
        positional=True,  # type: bool
    ):
        # type: (...) -> None
        if positional:
            parser.add_argument(
                "lockfile",
                nargs=1,
                help="The Pex lock file to {verb}".format(verb=verb),
            )
        else:
            parser.add_argument(
                "--lock",
                required=True,
                help="The Pex lock file to {verb}".format(verb=verb),
            )

    @classmethod
    def _add_lock_options(cls, parser):
        # type: (_ActionsContainer) -> None
        resolver_options.register_lock_options(parser)

    @classmethod
    def _add_create_arguments(cls, create_parser):
        # type: (_ActionsContainer) -> None
        cls.add_create_lock_options(create_parser)
        cls.add_output_option(create_parser, entity="lock")

    @classmethod
    def add_create_lock_options(cls, create_parser):
        create_parser.add_argument(
            "--style",
            default=LockStyle.STRICT,
            choices=LockStyle.values(),
            type=LockStyle.for_value,
            help=(
                "The style of lock to generate. The {strict!r} style is the default and generates "
                "a lock file that contains exactly the distributions that would be used in a local "
                "PEX build. If an sdist would be used, the sdist is included, but if a wheel would "
                "be used, an accompanying sdist will not be included. The {sources!r} style "
                "includes locks containing both wheels and the associated sdists when available. "
                "The {universal!r} style generates a universal lock for all possible target "
                "interpreters and platforms, although the scope can be constrained via one or more "
                "--interpreter-constraint. Of the three lock styles, only {strict!r} can give you "
                "full confidence in the lock since it includes exactly the artifacts that are "
                "included in the local PEX you'll build to test the lock result with before "
                "checking in the lock. With the other two styles you lock un-vetted artifacts in "
                "addition to the {strict!r} ones; so, even though you can be sure to reproducibly "
                "resolve those same un-vetted artifacts in the future, they're still un-vetted and "
                "could be innocently or maliciously different from the {strict!r} artifacts you "
                "can locally vet before committing the lock to version control. The effects of the "
                "differences could range from failing a resolve using the lock when the un-vetted "
                "artifacts have different dependencies from their sibling artifacts, to your "
                "application crashing due to different code in the sibling artifacts to being "
                "compromised by differing code in the sibling artifacts. So, although the more "
                "permissive lock styles will allow the lock to work on a wider range of machines /"
                "are apparently more convenient, the convenience comes with a potential price and "
                "using these styles should be considered carefully.".format(
                    strict=LockStyle.STRICT,
                    sources=LockStyle.SOURCES,
                    universal=LockStyle.UNIVERSAL,
                )
            ),
        )
        create_parser.add_argument(
            "--target-system",
            dest="target_systems",
            default=[],
            action="append",
            choices=TargetSystem.values(),
            type=TargetSystem.for_value,
            help=(
                "The target operating systems to generate the lock for. This option applies only "
                "to `--style {universal}` locks and restricts the locked artifacts to those "
                "compatible with the specified target operating systems. By default, {universal!r} "
                "style locks include artifacts for all operating systems.".format(
                    universal=LockStyle.UNIVERSAL,
                )
            ),
        )
        cls._add_lock_options(create_parser)
        cls._add_resolve_options(create_parser)
        cls.add_json_options(create_parser, entity="lock", include_switch=False)

    @classmethod
    def _add_export_arguments(
        cls,
        export_parser,  # type: _ActionsContainer
        lockfile_option_positional=True,  # type: bool
    ):
        # type: (...) -> None
        export_parser.add_argument(
            "--format",
            default=ExportFormat.PIP,
            choices=ExportFormat.values(),
            type=ExportFormat.for_value,
            help=(
                "The format to export the lock to. Currently only the Pip requirements file "
                "formats (using `--hash` or bare) are supported."
            ),
        )
        export_parser.add_argument(
            "--sort-by",
            default=ExportSortBy.SPECIFICITY,
            choices=ExportSortBy.values(),
            type=ExportSortBy.for_value,
            help="How to sort the requirements in the export (if supported).",
        )
        cls._add_lockfile_option(
            export_parser, verb="export", positional=lockfile_option_positional
        )
        cls._add_lock_options(export_parser)
        cls.add_output_option(export_parser, entity="lock")
        cls._add_target_options(export_parser)
        resolver_options_parser = cls._create_resolver_options_group(export_parser)
        resolver_options.register_network_options(resolver_options_parser)

    @classmethod
    def _add_export_subset_arguments(cls, export_subset_parser):
        # type: (_ActionsContainer) -> None
        cls._add_export_arguments(export_subset_parser, lockfile_option_positional=False)
        requirement_options.register(export_subset_parser)

    @classmethod
    def _add_update_arguments(cls, update_parser):
        # type: (_ActionsContainer) -> None
        update_parser.add_argument(
            "-p",
            "--project",
            "--update-project",
            dest="update_projects",
            action="append",
            default=[],
            type=str,
            help=(
                "Attempt to update these projects in the lock, leaving all others unchanged. "
                "If the projects aren't already in the lock, attempt to add them as top-level "
                "requirements leaving all others unchanged. If a project is already in the lock "
                "and is specified by a top-level requirement, the allowable updates will be "
                "constrained by the original top-level requirement. In other words, for an "
                "original top-level requirement of `requests>=2,<4` you could specify "
                "`-p requests` to pick up new releases in the `>=2,<4` range specified by the "
                "original top-level requirement or you could specify an overlapping range like "
                "`-p 'requests<3.6,!==3.8.2'` to, for example, downgrade away from a newly "
                "identified security vulnerability. If you specify a disjoint requirement like "
                "`-p 'requests>=4'`, the operation will fail. If you really want to replace the "
                "original top-level requirement and have this operation succeed, use the "
                "`--replace-project` option instead. This option is mutually exclusive with "
                "`--pin`."
            ),
        )
        update_parser.add_argument(
            "-R",
            "--replace-project",
            dest="replace_projects",
            action="append",
            default=[],
            type=str,
            help=(
                "Attempt to replace these projects in the lock, leaving all others unchanged. "
                "If the projects aren't already in the lock, attempt to add them as top-level "
                "requirements leaving all others unchanged. If a project is already in the lock "
                "and is specified by a top-level requirement, that top-level requirement will be "
                "replaced. This option is mutually exclusive with `--pin`."
            ),
        )
        update_parser.add_argument(
            "-d",
            "--delete-project",
            dest="delete_projects",
            action="append",
            default=[],
            type=str,
            help=(
                "Attempt to delete these projects from the lock, leaving all others unchanged. "
                "If the projects are not top-level requirements requested in the original lock or "
                "else they are but are transitive dependencies of the remaining top-level "
                "requirements, then no deletion will be performed in order to maintain lock "
                "integrity. This option is mutually exclusive with `--pin`."
            ),
        )
        cls.add_update_lock_options(update_parser)
        update_parser.add_argument(
            "--pin",
            action=HandleBoolAction,
            default=False,
            type=bool,
            help=(
                "When performing the update, pin all projects in the lock to their current "
                "versions. This is useful to pick up newly published wheels for those projects or "
                "else switch repositories from the original ones when used in conjunction with any "
                "of --index, --no-pypi and --find-links. When specifying `--pin`, it is an error "
                "to also specify lock modifications via `-p` / `--project`, "
                "`-R` / `--replace-project` or `-d` / `--delete-project`."
            ),
        )

        update_parser.add_argument(
            "--fingerprint-mismatch",
            default=FingerprintMismatch.ERROR,
            choices=FingerprintMismatch.values(),
            type=FingerprintMismatch.for_value,
            help=(
                "What to do when a lock update would result in at least one artifact fingerprint "
                "changing: {ignore!r} the mismatch and use the new fingerprint, {warn!r} about the "
                "mismatch but use the new fingerprint anyway or {error!r} and refuse to use the "
                "new mismatching fingerprint".format(
                    ignore=FingerprintMismatch.IGNORE,
                    warn=FingerprintMismatch.WARN,
                    error=FingerprintMismatch.ERROR,
                )
            ),
        )
        cls._add_lockfile_option(update_parser, verb="update")
        cls._add_lock_options(update_parser)
        cls.add_json_options(update_parser, entity="lock", include_switch=False)
        cls._add_target_options(update_parser)
        resolver_options_parser = cls._create_resolver_options_group(update_parser)
        resolver_options.register_repos_options(resolver_options_parser)
        resolver_options.register_network_options(resolver_options_parser)
        resolver_options.register_max_jobs_option(resolver_options_parser)
        resolver_options.register_use_pip_config(resolver_options_parser)

    @classmethod
    def add_update_lock_options(
        cls,
        update_parser,  # type: _ActionsContainer
        include_strict=True,  # type: bool
    ):
        if include_strict:
            update_parser.add_argument(
                "--strict",
                "--no-strict",
                "--non-strict",
                action=HandleBoolAction,
                default=True,
                type=bool,
                help=(
                    "Require all target platforms in the lock be updated at once. If any target "
                    "platform in the lock file does not have a representative local interpreter to "
                    "execute the lock update with, the update will fail."
                ),
            )
        update_parser.add_argument(
            "-n",
            "--dry-run",
            "--no-dry-run",
            action=HandleDryRunAction,
            help=(
                "Don't update the lock file; just report what updates would be made. By default, "
                "the report is to STDOUT and the exit code is zero. If a value of {check!r} is "
                "passed, the report is to STDERR and the exit code is non-zero.".format(
                    check=DryRunStyle.CHECK
                )
            ),
        )

    @classmethod
    def _add_sync_arguments(cls, sync_parser):
        # type: (_ActionsContainer) -> None
        sync_parser.add_argument(
            "--venv",
            help=(
                "Synchronize this venv to the contents of the lock after synchronizing the "
                "contents of the lock to the passed requirements."
            ),
        )
        sync_parser.add_argument(
            "--venv-python",
            help="Use this interpreter to create the `--venv` if the venv doesn't exist.",
        )
        sync_parser.add_argument(
            "--pip",
            "--retain-pip",
            "--no-pip",
            "--no-retain-pip",
            action=HandleBoolAction,
            default=False,
            help=(
                "When syncing a venv in the default `--no-pip`/`--no-retain-pip` mode, new venvs "
                "will be created without Pip installed in them and existing venvs will have Pip "
                "removed unless the lock being synced specifies a locked Pip, in which case that "
                "locked Pip version will be ensured. When syncing a venv in the "
                "`--pip`/`--retain-pip` mode, new venvs will be created with Pip installed in them "
                "and existing venvs will have Pip installed. The version of Pip installed will be "
                "taken from the lock being synced is present, and will be the latest compatible "
                "with the venv interpreter otherwise."
            ),
        )
        sync_parser.add_argument(
            "-y",
            "--yes",
            default=False,
            action="store_true",
            help=(
                "Do not prompt when deleting or overwriting venv distributions during a venv sync."
            ),
        )
        cls._add_lockfile_option(sync_parser, verb="sync", positional=False)
        cls.add_create_lock_options(sync_parser)
        cls.add_update_lock_options(sync_parser, include_strict=False)

    @classmethod
    def add_extra_arguments(
        cls,
        parser,  # type: ArgumentParser
    ):
        # type: (...) -> None
        subcommands = cls.create_subcommands(
            parser,
            description="PEX lock files can be operated on using any of the following subcommands.",
        )
        with subcommands.parser(
            name="create", help="Create a lock file.", func=cls._create
        ) as create_parser:
            cls._add_create_arguments(create_parser)
        with subcommands.parser(
            name="export",
            help="Export a Pex lock file for a single targeted environment in a different format.",
            func=cls._export,
        ) as export_parser:
            cls._add_export_arguments(export_parser)
        with subcommands.parser(
            name="export-subset",
            help=(
                "Export a subset of a Pex lock file for a single targeted environment in a "
                "different format."
            ),
            func=cls._export_subset,
        ) as export_subset_parser:
            cls._add_export_subset_arguments(export_subset_parser)
        with subcommands.parser(
            name="update", help="Update a Pex lock file.", func=cls._update
        ) as update_parser:
            cls._add_update_arguments(update_parser)
        with subcommands.parser(
            name="sync",
            help=(
                "Create or update a Pex lock file from requirements and optionally synchronize a "
                "venv to it."
            ),
            passthrough_args=(
                "The path to a venv directory to sync with this lock or else a command line whose "
                "1st argument resolves to a binary in a venv."
            ),
            func=cls._sync,
        ) as sync_parser:
            cls._add_sync_arguments(sync_parser)

    def _resolve_targets(
        self,
        action,  # type: str
        style,  # type: LockStyle.Value
        target_configuration=None,  # type: Optional[TargetConfiguration]
    ):
        # type: (...) -> Union[Targets, Error]

        target_config = target_configuration or target_options.configure(self.options)
        if style is not LockStyle.UNIVERSAL:
            return target_config.resolve_targets()

        if target_config.pythons:
            return Error(
                "When {action} a {universal} lock, the interpreters the resulting lock applies "
                "to can only be constrained via --interpreter-constraint. There {were} "
                "{num_pythons} --python specified.".format(
                    action=action,
                    universal=LockStyle.UNIVERSAL.value,
                    were="were" if len(target_config.pythons) > 1 else "was",
                    num_pythons=len(target_config.pythons),
                )
            )

        if not target_config.interpreter_constraints:
            return Targets(
                platforms=target_config.platforms,
                complete_platforms=target_config.complete_platforms,
                assume_manylinux=target_config.assume_manylinux,
            )

        try:
            interpreter = next(target_config.interpreter_configuration.iter_interpreters())
        except InterpreterConstraintsNotSatisfied as e:
            return Error(
                "When {action} a universal lock with an --interpreter-constraint, an "
                "interpreter matching the constraint must be found on the local system but "
                "none was: {err}".format(action=action, err=e)
            )
        return Targets(
            interpreters=(interpreter,),
            platforms=target_config.platforms,
            complete_platforms=target_config.complete_platforms,
            assume_manylinux=target_config.assume_manylinux,
        )

    def _gather_requirements(
        self,
        pip_configuration,  # type: PipConfiguration
        targets,  # type: Targets
    ):
        # type: (...) -> RequirementConfiguration
        requirement_configuration = requirement_options.configure(self.options)
        projects = project.get_projects(self.options)
        if not projects:
            return requirement_configuration

        requirements = OrderedSet(requirement_configuration.requirements)
        with TRACER.timed(
            "Collecting requirements from {count} local {projects}".format(
                count=len(projects), projects=pluralize(projects, "project")
            )
        ):
            requirements.update(
                str(req)
                for req in projects.collect_requirements(
                    resolver=ConfiguredResolver(pip_configuration),
                    interpreter=targets.interpreter,
                    pip_version=pip_configuration.version,
                    max_jobs=pip_configuration.max_jobs,
                )
            )
        return attr.evolve(requirement_configuration, requirements=requirements)

    def _create(self):
        # type: () -> Result
        target_configuration = target_options.configure(self.options)
        if self.options.style == LockStyle.UNIVERSAL:
            lock_configuration = LockConfiguration(
                style=LockStyle.UNIVERSAL,
                requires_python=tuple(
                    str(interpreter_constraint.requires_python)
                    for interpreter_constraint in target_configuration.interpreter_constraints
                ),
                target_systems=tuple(self.options.target_systems),
            )
        elif self.options.target_systems:
            return Error(
                "The --target-system option only applies to --style {universal} locks.".format(
                    universal=LockStyle.UNIVERSAL.value
                )
            )
        else:
            lock_configuration = LockConfiguration(style=self.options.style)

        targets = try_(
            self._resolve_targets(
                action="creating",
                style=self.options.style,
                target_configuration=target_configuration,
            )
        )
        pip_configuration = try_(
            finalize_resolve_config(
                resolver_configuration=resolver_options.create_pip_configuration(self.options),
                targets=targets,
                context="lock creation",
            )
        )
        requirement_configuration = self._gather_requirements(pip_configuration, targets)
        dependency_config = dependency_configuration.configure(self.options)
        self._dump_lockfile(
            try_(
                create(
                    lock_configuration=lock_configuration,
                    requirement_configuration=requirement_configuration,
                    targets=targets,
                    pip_configuration=pip_configuration,
                    dependency_configuration=dependency_config,
                )
            )
        )
        return Ok()

    def _load_lockfile(self):
        # type: () -> Tuple[str, Lockfile]
        lock_file_path = self.options.lock if "lock" in self.options else self.options.lockfile[0]
        return lock_file_path, try_(parse_lockfile(self.options, lock_file_path=lock_file_path))

    def _get_path_mappings(self):
        # type: () -> PathMappings
        return resolver_options.get_path_mappings(self.options)

    def _dump_lockfile(
        self,
        lock_file,  # type: Lockfile
        output=None,  # type: Optional[IO]
    ):
        # type: (...) -> None
        def dump_with_terminating_newline(out):
            # json.dump() does not write the newline terminating the last line, but some
            # JSON linters, and line-based tools in general, expect it, and since these
            # files are intended to be checked in to repos that may enforce this, we oblige.
            self.dump_json(
                self.options,
                json_codec.as_json_data(
                    lockfile=lock_file, path_mappings=self._get_path_mappings()
                ),
                out=out,
                sort_keys=True,
            )
            out.write("\n")

        if output:
            dump_with_terminating_newline(out=output)
        else:
            with self.output(self.options) as output:
                dump_with_terminating_newline(out=output)

    def _export(self, requirement_configuration=RequirementConfiguration()):
        # type: (RequirementConfiguration) -> Result
        supported_formats = ExportFormat.PIP, ExportFormat.PIP_NO_HASHES
        if self.options.format not in supported_formats:
            return Error(
                "Only the Pip lock formats are supported currently. "
                "Choose one of: {choices}".format(choices=" or ".join(map(str, supported_formats)))
            )

        lockfile_path, lock_file = self._load_lockfile()
        targets = target_options.configure(self.options).resolve_targets()
        target = targets.require_unique_target(
            purpose="exporting a lock in the {pip!r} format".format(pip=ExportFormat.PIP)
        )

        network_configuration = resolver_options.create_network_configuration(self.options)
        with TRACER.timed("Selecting locks for {target}".format(target=target)):
            subset_result = try_(
                subset(
                    targets=targets,
                    lock=lock_file,
                    requirement_configuration=requirement_configuration,
                    network_configuration=network_configuration,
                    build_configuration=lock_file.build_configuration(),
                    transitive=lock_file.transitive,
                    include_all_matches=True,
                )
            )

        if len(subset_result.subsets) != 1:
            resolved = Resolved.most_specific(
                resolved_subset.resolved for resolved_subset in subset_result.subsets
            )
            pex_warnings.warn(
                "Only a single lock can be exported in the {pip!r} format.\n"
                "There were {count} locks stored in {lockfile} that were applicable for the "
                "selected target: {target}; so using the most specific lock with platform "
                "{platform}.".format(
                    count=len(subset_result.subsets),
                    lockfile=lock_file.source,
                    pip=ExportFormat.PIP,
                    target=target,
                    platform=resolved.source.platform_tag,
                )
            )
        else:
            resolved = subset_result.subsets[0].resolved

        requirement_by_pin = {}  # type: Dict[Pin, str]
        fingerprints_by_pin = OrderedDict()  # type: OrderedDict[Pin, List[Fingerprint]]
        warnings = []  # type: List[str]

        def add_warning(
            type_,  # type: str
            requirement,  # type: str
        ):
            # type: (...) -> str
            warnings.append("{type} {requirement!r}".format(type=type_, requirement=requirement))
            return requirement

        for downloaded_artifact in resolved.downloadable_artifacts:
            if isinstance(downloaded_artifact.artifact, LocalProjectArtifact):
                requirement_by_pin[downloaded_artifact.pin] = add_warning(
                    "local project requirement",
                    requirement="{project_name} @ file://{directory}".format(
                        project_name=downloaded_artifact.pin.project_name,
                        directory=downloaded_artifact.artifact.directory,
                    ),
                )
            elif isinstance(downloaded_artifact.artifact, VCSArtifact):
                requirement_by_pin[downloaded_artifact.pin] = add_warning(
                    "VCS requirement",
                    requirement=downloaded_artifact.artifact.as_unparsed_requirement(
                        downloaded_artifact.pin.project_name
                    ),
                )
            else:
                requirement_by_pin[downloaded_artifact.pin] = "{project_name}=={version}".format(
                    project_name=downloaded_artifact.pin.project_name,
                    version=downloaded_artifact.pin.version.raw,
                )
            fingerprints_by_pin.setdefault(downloaded_artifact.pin, []).append(
                downloaded_artifact.artifact.fingerprint
            )

        if self.options.format is ExportFormat.PIP and warnings:
            print(
                "The requirements exported from {lockfile} include the following requirements\n"
                "that tools likely won't support --hash for:\n"
                "{warnings}\n"
                "\n"
                "If you can accept a lack of hash checking you can specify "
                "`--format pip-no-hashes`.\n".format(
                    lockfile=lockfile_path,
                    warnings="\n".join(
                        "+ {warning}".format(warning=warning) for warning in warnings
                    ),
                ),
                file=sys.stderr,
            )
        with self.output(self.options) as output:
            pins = fingerprints_by_pin.keys()  # type: Iterable[Pin]
            if self.options.sort_by == ExportSortBy.PROJECT_NAME:
                pins = sorted(pins, key=attrgetter("project_name.normalized"))
            for pin in pins:
                requirement = requirement_by_pin[pin]
                if self.options.format is ExportFormat.PIP_NO_HASHES:
                    print(requirement, file=output)
                else:
                    fingerprints = fingerprints_by_pin[pin]
                    output.write(
                        "{requirement} \\\n"
                        "  {hashes}\n".format(
                            requirement=requirement,
                            hashes=" \\\n  ".join(
                                "--hash={algorithm}:{hash}".format(
                                    algorithm=fingerprint.algorithm, hash=fingerprint.hash
                                )
                                for fingerprint in fingerprints
                            ),
                        )
                    )
        return Ok()

    def _export_subset(self):
        requirement_configuration = requirement_options.configure(self.options)
        return self._export(requirement_configuration=requirement_configuration)

    def _create_lock_update_request(
        self,
        lock_file_path,  # type: str
        lock_file,  # type: Lockfile
        dependency_config=DependencyConfiguration(),  # type: DependencyConfiguration
    ):
        # type: (...) -> Union[LockUpdateRequest, Error]

        network_configuration = resolver_options.create_network_configuration(self.options)
        lock_updater = LockUpdater.create(
            lock_file=lock_file,
            repos_configuration=resolver_options.create_repos_configuration(self.options),
            network_configuration=network_configuration,
            max_jobs=resolver_options.get_max_jobs_value(self.options),
            use_pip_config=resolver_options.get_use_pip_config_value(self.options),
            dependency_configuration=dependency_config,
        )

        target_configuration = target_options.configure(self.options)
        targets = try_(
            self._resolve_targets(
                action="updating", style=lock_file.style, target_configuration=target_configuration
            )
        )

        lock_updater = attr.evolve(
            lock_updater,
            pip_configuration=try_(
                finalize_resolve_config(
                    resolver_configuration=lock_updater.pip_configuration,
                    targets=targets,
                    context="lock updating",
                    pip_version=lock_file.pip_version,
                )
            ),
        )

        with TRACER.timed("Selecting locks to update"):
            locked_resolve_count = len(lock_file.locked_resolves)
            if lock_file.style is LockStyle.UNIVERSAL and locked_resolve_count != 1:
                return Error(
                    "The lock at {path} contains {count} locked resolves; so it "
                    "cannot be updated as a universal lock which requires exactly one locked "
                    "resolve.".format(path=lock_file_path, count=locked_resolve_count)
                )
            if locked_resolve_count == 1:
                locked_resolve = lock_file.locked_resolves[0]
                update_targets = (
                    [LocalInterpreter.create(targets.interpreter)]
                    if lock_file.style is LockStyle.UNIVERSAL
                    else targets.unique_targets()
                )
                update_requests = [
                    ResolveUpdateRequest(target=target, locked_resolve=locked_resolve)
                    for target in update_targets
                ]
            else:
                # N.B.: With 1 locked resolve in the lock file we're updating, there is no ambiguity
                # about which locked resolve should be paired with which target, but when there is
                # more than 1 locked resolve, we need to match up targets with the locked resolve
                # they should be responsible for updating. Here we use the existing subset logic
                # which is stricter than necessary in the sync case since it takes into account
                # artifacts. For any locked resolve with locked requirements that only have platform
                # specific wheel artifacts, this can prevent an update of a locked resolve to a new
                # Python version or platform.
                #
                # TODO(John Sirois): Consider implementing more permissive locked resolve selection
                #   logic to support syncing lock files containing multiple locked resolves:
                #     https://github.com/pex-tool/pex/issues/2386

                subset_result = try_(
                    subset(
                        targets=targets,
                        lock=lock_file,
                        network_configuration=network_configuration,
                        build_configuration=lock_file.build_configuration(),
                        transitive=lock_file.transitive,
                    )
                )
                update_requests = [
                    ResolveUpdateRequest(
                        target=resolved_subset.target,
                        locked_resolve=resolved_subset.resolved.source,
                    )
                    for resolved_subset in subset_result.subsets
                ]
        if getattr(self.options, "strict", False) and lock_file.style is not LockStyle.UNIVERSAL:
            missing_updates = set(lock_file.locked_resolves) - {
                update_request.locked_resolve for update_request in update_requests
            }
            if missing_updates:
                return Error(
                    "This lock update is --strict but the following platforms present in "
                    "{lock_file} were not found on the local machine:\n"
                    "{missing_platforms}\n"
                    "You might be able to correct this by adjusting target options like "
                    "--python-path or else by relaxing the update to be --non-strict.".format(
                        lock_file=lock_file_path,
                        missing_platforms="\n".join(
                            sorted(
                                "+ {platform}".format(platform=locked_resolve.platform_tag)
                                for locked_resolve in missing_updates
                            )
                        ),
                    )
                )

        return LockUpdateRequest(targets, lock_file_path, lock_updater, update_requests)

    def _process_lock_update(
        self,
        lock_update,  # type: LockUpdate
        lock_file,  # type: Lockfile
        lock_file_path,  # type: str
    ):
        # type: (...) -> Result

        original_requirements_by_project_name = OrderedDict(
            (requirement.project_name, requirement) for requirement in lock_file.requirements
        )
        requirements_by_project_name = OrderedDict(
            (requirement.project_name, requirement) for requirement in lock_update.requirements
        )

        original_constraints_by_project_name = OrderedDict(
            (constraint.project_name, constraint) for constraint in lock_file.constraints
        )
        constraints_by_project_name = original_constraints_by_project_name.copy()

        dry_run = self.options.dry_run
        path_mappings = self._get_path_mappings()
        output = sys.stdout if dry_run is DryRunStyle.DISPLAY else sys.stderr
        updates = []  # type: List[Union[DeleteUpdate, VersionUpdate, ArtifactsUpdate]]
        warnings = []  # type: List[str]
        for resolve_update in lock_update.resolves:
            platform = resolve_update.updated_resolve.target_platform
            if not resolve_update.updates:
                print(
                    "No updates for lock generated by {platform}.".format(platform=platform),
                    file=output,
                )
                continue

            print(
                "Updates for lock generated by {platform}:".format(platform=platform), file=output
            )
            fingerprint_updates = OrderedDict()  # type: OrderedDict[ProjectName, Version]
            for project_name, update in resolve_update.updates.items():
                if not update:
                    print(
                        "  There {tense} no updates for {project_name}".format(
                            tense="would be" if dry_run else "were",
                            project_name=project_name,
                        ),
                        file=output,
                    )
                    continue

                updates.append(update)
                if isinstance(update, DeleteUpdate):
                    print(
                        "  {lead_in} {project_name} {deleted_version}".format(
                            lead_in="Would delete" if dry_run else "Deleted",
                            project_name=project_name,
                            deleted_version=update.version,
                        ),
                        file=output,
                    )
                    production_assert(
                        project_name not in requirements_by_project_name,
                        "Deletes should have been unconditionally removed from requirements "
                        "earlier. Found deleted project {project_name} in updated requirements:\n"
                        "{requirements}".format(
                            project_name=project_name,
                            requirements="\n".join(map(str, requirements_by_project_name.values())),
                        ),
                    )
                    constraints_by_project_name.pop(project_name, None)
                elif isinstance(update, VersionUpdate):
                    update_req = lock_update.update_requirements_by_project_name.get(project_name)
                    if update.original:
                        print(
                            "  {lead_in} {project_name} from {original_version} to "
                            "{updated_version}".format(
                                lead_in="Would update" if dry_run else "Updated",
                                project_name=project_name,
                                original_version=update.original,
                                updated_version=update.updated,
                            ),
                            file=output,
                        )
                        # Only update the constraint if it is truly a constraint. If it's just the
                        # project name with no specifier, markers, etc., then it was just used to
                        # grab the latest version in the range already constrained by an existing
                        # requirement or constraint.
                        if update_req and str(update_req) != update_req.name:
                            constraints_by_project_name[project_name] = update_req.as_constraint()
                    else:
                        print(
                            "  {lead_in} {project_name} {updated_version}".format(
                                lead_in="Would add" if dry_run else "Added",
                                project_name=project_name,
                                updated_version=update.updated,
                            ),
                            file=output,
                        )
                        if update_req:
                            requirements_by_project_name[project_name] = update_req
                else:
                    message_lines = [
                        "  {lead_in} {project_name} {version} artifacts:".format(
                            lead_in="Would update" if dry_run else "Updated",
                            project_name=project_name,
                            version=update.version,
                        )
                    ]
                    if update.added:
                        message_lines.extend(
                            "    + {added}".format(
                                added=path_mappings.maybe_canonicalize(artifact.url.download_url)
                            )
                            for artifact in update.added
                        )
                    if update.updated:
                        if any(
                            isinstance(change, (FingerprintUpdate, ArtifactUpdate))
                            for change in update.updated
                        ):
                            fingerprint_updates[project_name] = update.version
                        message_lines.extend(
                            "    {changed}".format(
                                changed=path_mappings.maybe_canonicalize(change.render_update())
                            )
                            for change in update.updated
                        )
                    if update.removed:
                        message_lines.extend(
                            "    - {removed}".format(
                                removed=path_mappings.maybe_canonicalize(artifact.url.download_url)
                            )
                            for artifact in update.removed
                        )

                    print("\n".join(message_lines), file=output)
            if fingerprint_updates:
                warnings.append(
                    "Detected fingerprint changes in the following locked {projects} for lock "
                    "generated by {platform}!\n{ids}".format(
                        platform=platform,
                        projects=pluralize(fingerprint_updates, "project"),
                        ids="\n".join(
                            "{project_name} {version}".format(
                                project_name=project_name, version=version
                            )
                            for project_name, version in fingerprint_updates.items()
                        ),
                    )
                )

        def process_req_edit(
            original,  # type: Optional[Constraint]
            final,  # type: Optional[Constraint]
        ):
            # type: (...) -> None
            if not original:
                print(
                    "  {lead_in} {requirement!r}".format(
                        lead_in="Would add" if dry_run else "Added",
                        requirement=str(final),
                    ),
                    file=output,
                )
            elif not final:
                print(
                    "  {lead_in} {requirement!r}".format(
                        lead_in="Would delete" if dry_run else "Deleted",
                        requirement=str(original),
                    ),
                    file=output,
                )
            else:
                print(
                    "  {lead_in} {original!r} to {final!r}".format(
                        lead_in="Would update" if dry_run else "Updated",
                        original=str(original),
                        final=str(final),
                    ),
                    file=output,
                )

        def process_req_edits(
            requirement_type,  # type: str
            original,  # type: Mapping[ProjectName, Constraint]
            final,  # type: Mapping[ProjectName, Constraint]
        ):
            # type: (...) -> Tuple[Tuple[Optional[Constraint], Optional[Constraint]], ...]
            edits = []  # type: List[Tuple[Optional[Constraint], Optional[Constraint]]]
            for name, original_req in original.items():
                final_req = final.get(name)
                if final_req != original_req:
                    edits.append((original_req, final_req))
            for name, final_req in final.items():
                if name not in original:
                    edits.append((None, final_req))
            if not edits:
                return ()

            print(
                "Updates to lock input {requirement_types}:".format(
                    requirement_types=pluralize(2, requirement_type)
                ),
                file=output,
            )
            for original_requirement, final_requirement in edits:
                process_req_edit(original=original_requirement, final=final_requirement)
            return tuple(edits)

        requirement_edits = process_req_edits(
            "requirement", original_requirements_by_project_name, requirements_by_project_name
        )
        constraints_edits = process_req_edits(
            "constraint", original_constraints_by_project_name, constraints_by_project_name
        )

        if not any((updates, requirement_edits, constraints_edits)):
            return Ok()

        if warnings:
            if self.options.fingerprint_mismatch in (
                FingerprintMismatch.WARN,
                FingerprintMismatch.ERROR,
            ):
                message = "\n".join(warnings)
                if self.options.fingerprint_mismatch is FingerprintMismatch.ERROR:
                    return Error(message)
                pex_warnings.warn(message)

        if dry_run:
            return Error() if dry_run is DryRunStyle.CHECK else Ok()

        with open(lock_file_path, "w") as fp:
            self._dump_lockfile(
                lock_file=attr.evolve(
                    lock_file,
                    pex_version=__version__,
                    requirements=SortedTuple(requirements_by_project_name.values(), key=str),
                    constraints=SortedTuple(constraints_by_project_name.values(), key=str),
                    locked_resolves=SortedTuple(
                        resolve_update.updated_resolve for resolve_update in lock_update.resolves
                    ),
                ),
                output=fp,
            )
        return Ok()

    def _update(self):
        # type: () -> Result

        pin = getattr(self.options, "pin", False)
        if pin and any(
            (
                self.options.update_projects,
                self.options.replace_projects,
                self.options.delete_projects,
            )
        ):
            return Error(
                "When executing a `--pin`ed update, no `-p` / `--project`, "
                "`-R` / `--replace-project` or `-d` / `--delete-project` modifications are "
                "allowed."
            )

        update_requirements = []  # type: List[Requirement]
        for project in self.options.update_projects:
            try:
                update_requirements.append(Requirement.parse(project))
            except RequirementParseError as e:
                return Error(
                    "Failed to parse project requirement to update {project!r}: {err}".format(
                        project=project, err=e
                    )
                )

        replace_requirements = []  # type: List[Requirement]
        for project in self.options.replace_projects:
            try:
                replace_requirements.append(Requirement.parse(project))
            except RequirementParseError as e:
                return Error(
                    "Failed to parse replacement project requirement {project!r}: {err}".format(
                        project=project, err=e
                    )
                )

        try:
            delete_projects = tuple(
                ProjectName(project, validated=True) for project in self.options.delete_projects
            )
        except ProjectName.InvalidError as e:
            return Error("Failed to parse project name to delete: {err}".format(err=e))

        lock_file_path, lock_file = self._load_lockfile()
        lock_update_request = try_(
            self._create_lock_update_request(lock_file_path=lock_file_path, lock_file=lock_file)
        )
        lock_update = lock_update_request.update(
            updates=update_requirements,
            replacements=replace_requirements,
            deletes=delete_projects,
            artifacts_can_change=pin,
        )
        if isinstance(lock_update, Result):
            return lock_update

        return self._process_lock_update(lock_update, lock_file, lock_file_path)

    def _sync(self):
        # type: () -> Result

        resolver_configuration = cast(
            LockRepositoryConfiguration, resolver_options.configure(self.options)
        )
        production_assert(isinstance(resolver_configuration, LockRepositoryConfiguration))
        pip_configuration = resolver_configuration.pip_configuration
        dependency_config = dependency_configuration.configure(self.options)

        target_configuration = target_options.configure(self.options)
        if self.options.style == LockStyle.UNIVERSAL:
            lock_configuration = LockConfiguration(
                style=LockStyle.UNIVERSAL,
                requires_python=tuple(
                    str(interpreter_constraint.requires_python)
                    for interpreter_constraint in target_configuration.interpreter_constraints
                ),
                target_systems=tuple(self.options.target_systems),
            )
        elif self.options.target_systems:
            return Error(
                "The --target-system option only applies to --style {universal} locks.".format(
                    universal=LockStyle.UNIVERSAL.value
                )
            )
        else:
            lock_configuration = LockConfiguration(style=self.options.style)

        lock_file_path = self.options.lock
        if os.path.exists(lock_file_path):
            build_configuration = pip_configuration.build_configuration
            original_lock_file = try_(parse_lockfile(self.options, lock_file_path=lock_file_path))
            lock_file = attr.evolve(
                original_lock_file,
                style=lock_configuration.style,
                requires_python=SortedTuple(lock_configuration.requires_python),
                target_systems=SortedTuple(lock_configuration.target_systems),
                pip_version=pip_configuration.version,
                resolver_version=pip_configuration.resolver_version,
                allow_prereleases=pip_configuration.allow_prereleases,
                allow_wheels=build_configuration.allow_wheels,
                only_wheels=SortedTuple(build_configuration.only_wheels),
                allow_builds=build_configuration.allow_builds,
                only_builds=SortedTuple(build_configuration.only_builds),
                prefer_older_binary=build_configuration.prefer_older_binary,
                use_pep517=build_configuration.use_pep517,
                build_isolation=build_configuration.build_isolation,
                transitive=pip_configuration.transitive,
                excluded=SortedTuple(dependency_config.excluded),
                overridden=SortedTuple(dependency_config.all_overrides()),
            )
            lock_update_request = try_(
                self._create_lock_update_request(
                    lock_file_path=lock_file_path,
                    lock_file=lock_file,
                    dependency_config=dependency_config,
                )
            )
            requirement_configuration = self._gather_requirements(
                pip_configuration, lock_update_request.targets
            )
            lock_update = lock_update_request.sync(
                requirement_configuration=requirement_configuration,
            )
            if isinstance(lock_update, Result):
                return lock_update

            try_(self._process_lock_update(lock_update, lock_file, lock_file_path))
        else:
            targets = try_(
                self._resolve_targets(
                    action="creating",
                    style=self.options.style,
                    target_configuration=target_configuration,
                )
            )
            requirement_configuration = self._gather_requirements(pip_configuration, targets)
            lockfile = try_(
                create(
                    lock_configuration=lock_configuration,
                    requirement_configuration=requirement_configuration,
                    targets=targets,
                    pip_configuration=pip_configuration,
                    dependency_configuration=dependency_config,
                )
            )
            if self.options.dry_run:
                output = sys.stdout if self.options.dry_run is DryRunStyle.DISPLAY else sys.stderr
                for locked_resolve in lockfile.locked_resolves:
                    print(
                        "Would lock {count} {project} for platform {platform}:".format(
                            count=len(locked_resolve.locked_requirements),
                            project=pluralize(locked_resolve.locked_requirements, "project"),
                            platform=locked_resolve.target_platform,
                        ),
                        file=output,
                    )
                    for locked_requirement in locked_resolve.locked_requirements:
                        print(
                            "  {project_name} {version}".format(
                                project_name=locked_requirement.pin.project_name,
                                version=locked_requirement.pin.version,
                            ),
                            file=output,
                        )
            else:
                with safe_open(lock_file_path, "w") as fp:
                    self._dump_lockfile(lockfile, output=fp)

        sync_target = None  # type: Optional[SyncTarget]
        if self.options.venv:
            if os.path.exists(self.options.venv):
                try:
                    venv = Virtualenv(self.options.venv)
                except InvalidVirtualenvError as e:
                    return Error("The given --venv is not a valid venv: {err}".format(err=e))
            else:
                if self.options.venv_python:
                    interpreter = (
                        PythonInterpreter.from_binary(self.options.venv_python)
                        if os.path.isfile(self.options.venv_python)
                        else PythonInterpreter.from_env(self.options.venv_python)
                    )
                else:
                    targets = target_options.configure(self.options).resolve_targets()
                    interpreters = [
                        target.get_interpreter()
                        for target in targets.unique_targets()
                        if isinstance(target, LocalInterpreter)
                    ]
                    if len(interpreters) != 1:
                        return Error(
                            "In order to create the venv {venv} the sync operation needs to know "
                            "which interpreter to use to create it.\n"
                            "Use `--venv-interpreter` to select an interpreter for venv creation."
                        )
                    interpreter = interpreters[0]
                venv = Virtualenv.create(self.options.venv, interpreter=interpreter)
            sync_target = SyncTarget.resolve_command(venv=venv, command=self.passthrough_args)
        elif self.passthrough_args:
            sync_target = try_(
                SyncTarget.resolve_venv(self.passthrough_args[0], *self.passthrough_args[1:])
            )
        if not sync_target:
            return Ok()

        if self.options.dry_run:
            would_update_venv_msg = "Would sync venv at {venv}".format(
                venv=sync_target.venv.venv_dir
            )
            if sync_target.command:
                return Ok(
                    os.linesep.join(
                        (
                            would_update_venv_msg + " and run the following command in it:",
                            "  " + " ".join(shlex_quote(arg) for arg in sync_target.command),
                        )
                    )
                )
            else:
                return Ok(would_update_venv_msg + ".")

        lock_file = try_(parse_lockfile(self.options, lock_file_path=lock_file_path))
        target = LocalInterpreter.create(sync_target.venv.interpreter)
        resolve_result = try_(
            resolve_from_lock(
                targets=Targets.from_target(target),
                lock=lock_file,
                resolver=ConfiguredResolver(pip_configuration),
                requirements=requirement_configuration.requirements,
                requirement_files=requirement_configuration.requirement_files,
                constraint_files=requirement_configuration.constraint_files,
                indexes=pip_configuration.repos_configuration.indexes,
                find_links=pip_configuration.repos_configuration.find_links,
                resolver_version=pip_configuration.resolver_version,
                network_configuration=pip_configuration.network_configuration,
                password_entries=pip_configuration.repos_configuration.password_entries,
                build_configuration=pip_configuration.build_configuration,
                transitive=pip_configuration.transitive,
                max_parallel_jobs=pip_configuration.max_jobs,
                pip_version=pip_configuration.version,
                use_pip_config=pip_configuration.use_pip_config,
                extra_pip_requirements=pip_configuration.extra_requirements,
                result_type=InstallableType.INSTALLED_WHEEL_CHROOT,
            )
        )
        return sync_target.sync(
            distributions=[
                resolved_distribution.distribution
                for resolved_distribution in resolve_result.distributions
                if resolved_distribution.target == target
            ],
            confirm=not self.options.yes,
            retain_pip=self.options.pip,
        )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import glob
import json
import os.path
import re
import subprocess
import sys
from typing import Optional

import pytest

from pex.common import is_exe
from pex.layout import Layout
from pex.orderedset import OrderedSet
from pex.scie import SciePlatform, ScieStyle
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from testing import IS_PYPY, PY_VER, make_env, run_pex_command

if TYPE_CHECKING:
    from typing import Any, Iterable, List


@pytest.mark.parametrize(
    "scie_style", [pytest.param(style, id=str(style)) for style in ScieStyle.values()]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=str(layout)) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="ZIPAPP"),
        pytest.param(["--venv"], id="VENV"),
        pytest.param(["--sh-boot"], id="ZIPAPP-sh-boot"),
        pytest.param(["--venv", "--sh-boot"], id="VENV-sh-boot"),
    ],
)
def test_basic(
    tmpdir,  # type: Any
    scie_style,  # type: ScieStyle.Value
    layout,  # type: Layout.Value
    execution_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "cowsay.pex")
    result = run_pex_command(
        args=[
            "cowsay==5.0",
            "-c",
            "cowsay",
            "-o",
            pex,
            "--scie",
            str(scie_style),
            "--layout",
            str(layout),
        ]
        + execution_mode_args
    )
    if PY_VER < (3, 8) or IS_PYPY:
        result.assert_failure(
            expected_error_re=r".*^{message}$".format(
                message=re.escape(
                    "You selected `--scie {style}`, but none of the selected targets have "
                    "compatible interpreters that can be embedded to form a scie:\n"
                    "{target}".format(
                        style=scie_style, target=LocalInterpreter.create().render_description()
                    )
                )
            ),
            re_flags=re.DOTALL | re.MULTILINE,
        )
        return
    if PY_VER >= (3, 13):
        result.assert_failure(
            expected_error_re=(
                r".*"
                r"^Failed to build 1 scie:$"
                r".*"
                r"^Provider: No released assets found for release [0-9]{{8}} Python {version} "
                r"of flavor install_only\.$".format(version=".".join(map(str, PY_VER)))
            ),
            re_flags=re.DOTALL | re.MULTILINE,
        )
        return
    result.assert_success()

    scie = os.path.join(str(tmpdir), "cowsay")
    assert b"| PAR! |" in subprocess.check_output(args=[scie, "PAR!"], env=make_env(PATH=None))


def test_multiple_platforms(tmpdir):
    # type: (Any) -> None

    def create_scies(
        output_dir,  # type: str
        extra_args=(),  # type: Iterable[str]
    ):
        pex = os.path.join(output_dir, "cowsay.pex")
        run_pex_command(
            args=[
                "cowsay==5.0",
                "-c",
                "cowsay",
                "-o",
                pex,
                "--scie",
                "lazy",
                "--platform",
                "linux-aarch64-cp-39-cp39",
                "--platform",
                "linux-x86_64-cp-310-cp310",
                "--platform",
                "macosx-10.9-arm64-cp-311-cp311",
                "--platform",
                "macosx-10.9-x86_64-cp-312-cp312",
            ]
            + list(extra_args)
        ).assert_success()

    python_version_by_platform = {
        SciePlatform.LINUX_AARCH64: "3.9",
        SciePlatform.LINUX_X86_64: "3.10",
        SciePlatform.MACOS_AARCH64: "3.11",
        SciePlatform.MACOS_X86_64: "3.12",
    }
    assert SciePlatform.CURRENT in python_version_by_platform

    def assert_platforms(
        output_dir,  # type: str
        expected_platforms,  # type: Iterable[SciePlatform.Value]
    ):
        # type: (...) -> None

        all_output_files = set(
            path
            for path in os.listdir(output_dir)
            if os.path.isfile(os.path.join(output_dir, path))
        )
        for platform in OrderedSet(expected_platforms):
            python_version = python_version_by_platform[platform]
            binary = platform.qualified_binary_name("cowsay")
            assert binary in all_output_files
            all_output_files.remove(binary)
            scie = os.path.join(output_dir, binary)
            assert is_exe(scie), "Expected --scie build to produce a {binary} binary.".format(
                binary=binary
            )
            if platform is SciePlatform.CURRENT:
                assert b"| PEX-scie wabbit! |" in subprocess.check_output(
                    args=[scie, "PEX-scie wabbit!"], env=make_env(PATH=None)
                )
                assert (
                    python_version
                    == subprocess.check_output(
                        args=[
                            scie,
                            "-c",
                            "import sys; print('.'.join(map(str, sys.version_info[:2])))",
                        ],
                        env=make_env(PEX_INTERPRETER=1),
                    )
                    .decode("utf-8")
                    .strip()
                )
        assert {"cowsay.pex"} == all_output_files, (
            "Expected one output scie for each platform plus the original cowsay.pex. All expected "
            "scies were found, but the remaining files are: {remaining_files}".format(
                remaining_files=all_output_files
            )
        )

    all_platforms_output_dir = os.path.join(str(tmpdir), "all-platforms")
    create_scies(output_dir=all_platforms_output_dir)
    assert_platforms(
        output_dir=all_platforms_output_dir,
        expected_platforms=(
            SciePlatform.LINUX_AARCH64,
            SciePlatform.LINUX_X86_64,
            SciePlatform.MACOS_AARCH64,
            SciePlatform.MACOS_X86_64,
        ),
    )

    # Now restrict the PEX's implied natural platform set of 4 down to 2 or 3 using
    # `--scie-platform`.
    restricted_platforms_output_dir = os.path.join(str(tmpdir), "restricted-platforms")
    create_scies(
        output_dir=restricted_platforms_output_dir,
        extra_args=[
            "--scie-platform",
            "current",
            "--scie-platform",
            str(SciePlatform.LINUX_AARCH64),
            "--scie-platform",
            str(SciePlatform.LINUX_X86_64),
        ],
    )
    assert_platforms(
        output_dir=restricted_platforms_output_dir,
        expected_platforms=(
            SciePlatform.CURRENT,
            SciePlatform.LINUX_AARCH64,
            SciePlatform.LINUX_X86_64,
        ),
    )


PRINT_VERSION_SCRIPT = "import sys; print('.'.join(map(str, sys.version_info[:3])))"


skip_if_pypy = pytest.mark.skipif(IS_PYPY, reason="PyPy targeted PEXes do not support --scie.")


@skip_if_pypy
def test_specified_interpreter(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "empty.pex")
    run_pex_command(
        args=[
            "-o",
            pex,
            "--scie",
            "lazy",
            # We pick a specific version that is not in the latest release but is known to provide
            # distributions for all platforms Pex tests run on.
            "--scie-pbs-release",
            "20221002",
            "--scie-python-version",
            "3.10.7",
        ],
    ).assert_success()

    assert (
        ".".join(map(str, sys.version_info[:3]))
        == subprocess.check_output(args=[pex, "-c", PRINT_VERSION_SCRIPT]).decode("utf-8").strip()
    )

    scie = os.path.join(str(tmpdir), "empty")
    assert b"3.10.7\n" == subprocess.check_output(args=[scie, "-c", PRINT_VERSION_SCRIPT])


@skip_if_pypy
def test_specified_science_binary(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    scie = os.path.join(str(tmpdir), "cowsay")
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "cowsay==6.0",
            "-c",
            "cowsay",
            "--scie",
            "lazy",
            "--scie-python-version",
            "3.12",
            "-o",
            scie,
            "--scie-science-binary",
            # N.B.: This custom version is both lower than the latest available version (0.4.2
            # at the time of writing) and higher than the minimum supported version of 0.3.0; so
            # we can prove we downloaded the custom version via this URL by checking the version
            # below since our next floor bump will be from 0.3.0 to at least 0.4.3.
            "https://github.com/a-scie/lift/releases/download/v0.4.0/{binary}".format(
                binary=SciePlatform.CURRENT.qualified_binary_name("science")
            ),
        ],
        env=make_env(PATH=None),
    ).assert_success()

    assert b"| Alternative SCIENCE Facts! |" in subprocess.check_output(
        args=[scie, "-t", "Alternative SCIENCE Facts!"]
    )

    science_binaries = glob.glob(os.path.join(pex_root, "scies", "science", "*", "bin", "science"))
    assert 1 == len(science_binaries)
    science = science_binaries[0]
    assert "0.4.0" == subprocess.check_output(args=[science, "--version"]).decode("utf-8").strip()


@skip_if_pypy
def test_custom_lazy_urls(tmpdir):
    # type: (Any) -> None

    scie = os.path.join(str(tmpdir), "empty")
    run_pex_command(
        args=[
            "-o",
            scie,
            "--scie",
            "lazy",
            "--scie-pbs-release",
            "20221002",
            "--scie-python-version",
            "3.10.7",
        ],
    ).assert_success()

    assert b"3.10.7\n" == subprocess.check_output(args=[scie, "-c", PRINT_VERSION_SCRIPT])

    pex_bootstrap_urls = os.path.join(str(tmpdir), "pex_bootstrap_urls.json")

    def make_20221002_3_10_7_file(platform):
        # type: (str) -> str
        return "cpython-3.10.7+20221002-{platform}-install_only.tar.gz".format(platform=platform)

    def make_20240415_3_10_14_url(platform):
        # type: (str) -> str
        return (
            "https://github.com/indygreg/python-build-standalone/releases/download/20240415/"
            "cpython-3.10.14+20240415-{platform}-install_only.tar.gz".format(platform=platform)
        )

    with open(pex_bootstrap_urls, "w") as fp:
        json.dump(
            {
                "ptex": {
                    make_20221002_3_10_7_file(platform): make_20240415_3_10_14_url(platform)
                    for platform in (
                        "aarch64-apple-darwin",
                        "x86_64-apple-darwin",
                        "aarch64-unknown-linux-gnu",
                        "x86_64-unknown-linux-gnu",
                    )
                }
            },
            fp,
        )

    process = subprocess.Popen(
        args=[scie, "-c", PRINT_VERSION_SCRIPT],
        env=make_env(
            PEX_BOOTSTRAP_URLS=pex_bootstrap_urls, SCIE_BASE=os.path.join(str(tmpdir), "nce")
        ),
        stderr=subprocess.PIPE,
    )
    _, stderr = process.communicate()
    assert 0 != process.returncode, (
        "Expected PEX_BOOTSTRAP_URLS to be used and the resulting fetched interpreter distribution "
        "to fail its digest check."
    )

    expected_platform = None  # type: Optional[str]
    if SciePlatform.CURRENT is SciePlatform.LINUX_AARCH64:
        expected_platform = "aarch64-unknown-linux-gnu"
    elif SciePlatform.CURRENT is SciePlatform.LINUX_X86_64:
        expected_platform = "x86_64-unknown-linux-gnu"
    elif SciePlatform.CURRENT is SciePlatform.MACOS_AARCH64:
        expected_platform = "aarch64-apple-darwin"
    elif SciePlatform.CURRENT is SciePlatform.MACOS_X86_64:
        expected_platform = "x86_64-apple-darwin"
    assert expected_platform is not None

    assert re.match(
        r"^.*Population of work directory failed: The tar\.gz destination .*{expected_file_name} "
        r"of size \d+ had unexpected hash: [a-f0-9]{{64}}$.*".format(
            expected_file_name=re.escape(make_20221002_3_10_7_file(expected_platform))
        ),
        stderr.decode("utf-8"),
        flags=re.DOTALL | re.MULTILINE,
    ), stderr.decode("utf-8")



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import glob
import os.path
from email.message import Message
from typing import Callable, Text, Tuple

from pex.common import safe_mkdtemp
from pex.dist_metadata import DistMetadata, DistMetadataFile, MetadataFiles, MetadataType
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Optional


def create_dist_metadata(
    project_name,  # type: str
    version,  # type: str
    requires_python=None,  # type: Optional[str]
    requires_dists=(),  # type: Iterable[str]
    location=None,  # type: Optional[str]
    metadata_type=MetadataType.DIST_INFO,  # type: MetadataType.Value
):
    # type: (...) -> DistMetadata

    pkg_info = Message()
    pkg_info.add_header("Name", project_name)
    pkg_info.add_header("Version", version)
    if requires_python:
        pkg_info.add_header("Requires-Python", requires_python)
    for requirement in requires_dists:
        pkg_info.add_header("Requires-Dist", requirement)

    resolved_location = location or safe_mkdtemp()
    metadata_dir = "{project_name}-{version}.{suffix}".format(
        project_name=project_name,
        version=version,
        suffix="dist-info" if metadata_type is MetadataType.DIST_INFO else "egg-info",
    )
    metadata_file_name = "METADATA" if metadata_type is MetadataType.DIST_INFO else "PKG-INFO"

    additional_metadata_files = ()  # type: Tuple[Text, ...]
    read_function = None  # type: Optional[Callable[[Text], bytes]]
    if os.path.isdir(resolved_location):
        additional_metadata_files = tuple(
            os.path.relpath(metadata_path, resolved_location)
            for metadata_path in glob.glob(os.path.join(resolved_location, metadata_dir, "*"))
            if os.path.basename(metadata_path) != metadata_file_name
        )
        if additional_metadata_files:

            def read_function(rel_path):
                # type: (Text) -> bytes
                with open(os.path.join(resolved_location, rel_path), "rb") as fp:
                    return fp.read()

    return DistMetadata.from_metadata_files(
        MetadataFiles(
            metadata=DistMetadataFile(
                type=metadata_type,
                location=resolved_location,
                rel_path=os.path.join(metadata_dir, metadata_file_name),
                project_name=ProjectName(project_name),
                version=Version(version),
                pkg_info=pkg_info,
            ),
            additional_metadata_files=additional_metadata_files,
            read_function=read_function,
        )
    )



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from textwrap import dedent

import pytest
from colors import colors

from pex.typing import TYPE_CHECKING
from testing import IntegResults, run_pex_command

if TYPE_CHECKING:
    from typing import Any, List


# N.B.: To test that we are running a REPL we'll show that it computes result even after
# assertion errors and exceptions that would normally halt a script.
# To check that it forwards python options we use -O to deactivate asserts
# See: https://docs.python.org/3/using/cmdline.html#cmdoption-O
@pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="ZIPAPP"),
        pytest.param(["--venv"], id="VENV"),
    ],
)
def test_repl_python_options(
    execution_mode_args,  # type: List[str]
    tmpdir,  # type: Any
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["ansicolors==1.1.8", "-o", pex] + execution_mode_args).assert_success()

    repl_commands = dedent(
        """
        import colors
        assert False
        raise Exception("customexc")
        result = 20 + 103
        print(colors.green("Worked: {}".format(result)))
        quit()
        """
    )

    def execute_pex(disable_assertions):
        # type: (bool) -> IntegResults
        args = [pex]
        if disable_assertions:
            args.append("-O")
        process = subprocess.Popen(
            args=args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate(input=repl_commands.encode("utf-8"))
        return IntegResults(
            output=stdout.decode("utf-8"),
            error=stderr.decode("utf-8"),
            return_code=process.returncode,
        )

    # The assertion will fail and print but since it is a REPL it will keep going
    # and compute the result
    result = execute_pex(disable_assertions=False)
    result.assert_success()
    assert "InteractiveConsole" in result.error
    assert "AssertionError" in result.error
    assert "customexc" in result.error
    assert colors.green("Worked: 123") in result.output

    # The -O will disable the assertion, but the regular exception will still get raised.
    result = execute_pex(disable_assertions=True)
    result.assert_success()
    assert "InteractiveConsole" in result.error
    assert "AssertionError" not in result.error
    assert "customexc" in result.error
    assert colors.green("Worked: 123") in result.output



# Copyright 2014 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys

TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import List, NoReturn, Optional, Tuple


if sys.version_info >= (3, 10):

    def orig_argv():
        # type: () -> List[str]
        return sys.orig_argv

else:
    try:
        import ctypes

        # N.B.: None of the PyPy versions we support <3.10 supports the pythonapi.
        from ctypes import pythonapi

        def orig_argv():
            # type: () -> List[str]

            # Under MyPy for Python 3.5, ctypes.POINTER is incorrectly typed. This code is tested
            # to work correctly in practice on all Pythons Pex supports.
            argv = ctypes.POINTER(  # type: ignore[call-arg]
                ctypes.c_char_p if sys.version_info[0] == 2 else ctypes.c_wchar_p
            )()

            argc = ctypes.c_int()
            pythonapi.Py_GetArgcArgv(ctypes.byref(argc), ctypes.byref(argv))

            # Under MyPy for Python 3.5, argv[i] has its type incorrectly evaluated. This code
            # is tested to work correctly in practice on all Pythons Pex supports.
            return [argv[i] for i in range(argc.value)]  # type: ignore[misc]

    except ImportError:
        # N.B.: This handles the older PyPy case.
        def orig_argv():
            # type: () -> List[str]
            return []


def __re_exec__(
    python,  # type: str
    python_args,  # type: List[str]
    *extra_python_args  # type: str
):
    # type: (...) -> NoReturn

    argv = [python]
    argv.extend(python_args)
    argv.extend(extra_python_args)

    os.execv(python, argv + sys.argv[1:])


__SHOULD_EXECUTE__ = __name__ == "__main__"


def __entry_point_from_filename__(filename):
    # type: (str) -> str

    # Either the entry point is "__main__" and we're in execute mode or "__pex__/__init__.py"
    # and we're in import hook mode.
    entry_point = os.path.dirname(filename)
    if __SHOULD_EXECUTE__:
        return entry_point
    return os.path.dirname(entry_point)


__INSTALLED_FROM__ = "__PEX_EXE__"


def __ensure_pex_installed__(
    pex,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
    python_args,  # type: List[str]
):
    # type: (...) -> Optional[str]

    from pex.layout import ensure_installed
    from pex.tracer import TRACER

    installed_location = ensure_installed(pex=pex, pex_root=pex_root, pex_hash=pex_hash)
    if not __SHOULD_EXECUTE__ or pex == installed_location:
        return installed_location

    # N.B.: This is read upon re-exec below to point sys.argv[0] back to the original pex
    # before unconditionally scrubbing the env var and handing off to user code.
    os.environ[__INSTALLED_FROM__] = pex

    TRACER.log(
        "Executing installed PEX for {pex} at {installed_location}".format(
            pex=pex, installed_location=installed_location
        )
    )
    __re_exec__(sys.executable, python_args, installed_location)


def __maybe_run_venv__(
    pex,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
    has_interpreter_constraints,  # type: bool
    hermetic_venv_scripts,  # type: bool
    pex_path,  # type: Tuple[str, ...]
    python_args,  # type: List[str]
):
    # type: (...) -> Optional[str]

    from pex.common import is_exe
    from pex.tracer import TRACER
    from pex.variables import venv_dir

    venv_root_dir = venv_dir(
        pex_file=pex,
        pex_root=pex_root,
        pex_hash=pex_hash,
        has_interpreter_constraints=has_interpreter_constraints,
        pex_path=pex_path,
    )
    venv_pex = os.path.join(venv_root_dir, "pex")
    if not __SHOULD_EXECUTE__ or not is_exe(venv_pex):
        # Code in bootstrap_pex will (re)create the venv after selecting the correct
        # interpreter.
        return venv_root_dir

    TRACER.log("Executing venv PEX for {pex} at {venv_pex}".format(pex=pex, venv_pex=venv_pex))
    venv_python = os.path.join(venv_root_dir, "bin", "python")
    if hermetic_venv_scripts:
        __re_exec__(venv_python, python_args, "-sE", venv_pex)
    else:
        __re_exec__(venv_python, python_args, venv_pex)


def boot(
    bootstrap_dir,  # type: str
    pex_root,  # type: str
    pex_hash,  # type: str
    has_interpreter_constraints,  # type: bool
    hermetic_venv_scripts,  # type: bool
    pex_path,  # type: Tuple[str, ...]
    is_venv,  # type: bool
    inject_python_args,  # type: Tuple[str, ...]
):
    # type: (...) -> int

    entry_point = None  # type: Optional[str]
    __file__ = globals().get("__file__")
    __loader__ = globals().get("__loader__")
    if __file__ is not None and os.path.exists(__file__):
        entry_point = __entry_point_from_filename__(__file__)
    elif __loader__ is not None:
        if hasattr(__loader__, "archive"):
            entry_point = __loader__.archive
        elif hasattr(__loader__, "get_filename"):
            # The source of the loader interface has changed over the course of Python history
            # from `pkgutil.ImpLoader` to `importlib.abc.Loader`, but the existence and
            # semantics of `get_filename` has remained constant; so we just check for the
            # method.
            entry_point = __entry_point_from_filename__(__loader__.get_filename())

    if entry_point is None:
        sys.stderr.write("Could not launch python executable!\\n")
        return 2

    python_args = list(inject_python_args)  # type: List[str]
    orig_args = orig_argv()
    if orig_args:
        for index, arg in enumerate(orig_args[1:], start=1):
            if os.path.exists(arg) and os.path.samefile(entry_point, arg):
                python_args.extend(orig_args[1:index])
                break

    installed_from = os.environ.pop(__INSTALLED_FROM__, None)
    sys.argv[0] = installed_from or sys.argv[0]

    sys.path[0] = os.path.abspath(sys.path[0])
    sys.path.insert(0, os.path.abspath(os.path.join(entry_point, bootstrap_dir)))

    venv_dir = None  # type: Optional[str]
    if not installed_from:
        os.environ["PEX"] = os.path.realpath(entry_point)
        from pex.variables import ENV, Variables

        pex_root = Variables.PEX_ROOT.value_or(ENV, pex_root)

        if not ENV.PEX_TOOLS and Variables.PEX_VENV.value_or(ENV, is_venv):
            venv_dir = __maybe_run_venv__(
                pex=entry_point,
                pex_root=pex_root,
                pex_hash=pex_hash,
                has_interpreter_constraints=has_interpreter_constraints,
                hermetic_venv_scripts=hermetic_venv_scripts,
                pex_path=ENV.PEX_PATH or pex_path,
                python_args=python_args,
            )
        entry_point = __ensure_pex_installed__(
            pex=entry_point, pex_root=pex_root, pex_hash=pex_hash, python_args=python_args
        )
        if entry_point is None:
            # This means we re-exec'd ourselves already; so this just appeases type checking.
            return 0
    else:
        os.environ["PEX"] = os.path.realpath(installed_from)

    from pex.pex_bootstrapper import bootstrap_pex

    bootstrap_pex(
        entry_point, python_args=python_args, execute=__SHOULD_EXECUTE__, venv_dir=venv_dir
    )
    return 0



# Copyright 2014 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import sys
import threading
import time
from contextlib import contextmanager

from pex.typing import TYPE_CHECKING, cast
from pex.variables import ENV

__all__ = ("TRACER", "TraceLogger")

if TYPE_CHECKING:
    from typing import IO, Any, Callable, Iterator, List, Optional


class Trace(object):
    __slots__ = ("msg", "verbosity", "parent", "children", "_clock", "_start", "_stop")

    def __init__(self, msg, parent=None, verbosity=1, clock=time):
        # type: (str, Optional[Trace], int, Any) -> None
        self.msg = msg
        self.verbosity = verbosity
        self.parent = parent
        if parent is not None:
            parent.children.append(self)
        self.children = []  # type: List[Trace]
        self._clock = clock
        self._start = cast(float, self._clock.time())
        self._stop = cast("Optional[float]", None)

    def stop(self):
        # type: () -> None
        self._stop = self._clock.time()

    def duration(self):
        # type: () -> float
        assert self._stop is not None
        return self._stop - self._start


class TraceLogger(object):
    """A multi-threaded tracer."""

    def __init__(self, predicate=None, output=sys.stderr, clock=time, prefix=""):
        # type: (Optional[Callable[[int], bool]], IO, Any, str) -> None
        """If predicate specified, it should take a "verbosity" integer and determine whether or not
        to log, e.g.

          def predicate(verbosity):
            try:
              return verbosity < int(os.environ.get('APP_VERBOSITY', 0))
            except ValueError:
              return False

        output defaults to sys.stderr, but can take any file-like object.
        """
        self._predicate = predicate or (lambda verbosity: True)
        self._length = cast("Optional[int]", None)
        self._output = output
        self._isatty = getattr(output, "isatty", False) and output.isatty()
        self._lock = threading.RLock()
        self._local = threading.local()
        self._clock = clock
        self._prefix = prefix

    def should_log(self, V):
        # type: (int) -> bool
        return self._predicate(V)

    def log(self, msg, V=1, end="\n"):
        # type: (str, int, str) -> None
        if not self.should_log(V):
            return
        if not self._isatty and end == "\r":
            # force newlines if we're not a tty
            end = "\n"
        trailing_whitespace = ""
        with self._lock:
            if self._length and self._length > (len(self._prefix) + len(msg)):
                trailing_whitespace = " " * (self._length - len(msg) - len(self._prefix))
            self._output.write("".join([self._prefix, msg, trailing_whitespace, end]))
            self._output.flush()
            self._length = (len(self._prefix) + len(msg)) if end == "\r" else 0

    def _print_trace_snippet(self, node):
        # type: (Trace) -> None
        node_verbosity = node.verbosity
        if not self.should_log(node_verbosity):
            return
        traces = []
        parent = node  # type: Optional[Trace]
        while parent:
            if self.should_log(parent.verbosity):
                traces.append(parent.msg)
            parent = parent.parent
        self.log(" :: ".join(reversed(traces)), V=node_verbosity, end="\r")

    def _print_trace(self, node, indent=0):
        # type: (Trace, int) -> None
        with self._lock:
            self.log(
                " " * indent + ("%s: %.1fms" % (node.msg, 1000.0 * node.duration())),
                V=node.verbosity,
            )
            for child in node.children:
                self._print_trace(indent=indent + 2, node=child)

    @contextmanager
    def timed(self, msg, V=1):
        # type: (str, int) -> Iterator[None]
        parent = Trace(
            msg, parent=getattr(self._local, "parent", None), verbosity=V, clock=self._clock
        )
        self._local.parent = parent
        self._print_trace_snippet(parent)
        try:
            yield
        finally:
            parent.stop()
            if parent.parent is not None:
                self._local.parent = parent.parent
            else:
                self._print_trace(parent)
                self._local.parent = None


TRACER = TraceLogger(
    predicate=lambda verbosity: verbosity <= ENV.PEX_VERBOSE,
    prefix="pex: ",
)



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import hashlib

from pex import hashing
from pex.compatibility import url_unquote, urlparse
from pex.dist_metadata import ProjectNameAndVersion, Requirement
from pex.hashing import HashlibHasher
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.requirements import ArchiveScheme, VCSScheme, parse_scheme
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import BinaryIO, Iterator, Mapping, Optional, Sequence, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Pin(object):
    @classmethod
    def canonicalize(cls, project_name_and_version):
        # type: (ProjectNameAndVersion) -> Pin
        return cls(
            project_name=ProjectName(project_name_and_version.project_name),
            version=Version(project_name_and_version.version),
        )

    project_name = attr.ib()  # type: ProjectName
    version = attr.ib()  # type: Version

    def as_requirement(self):
        # type: () -> Requirement
        return Requirement.parse(
            "{project_name}=={version}".format(project_name=self.project_name, version=self.version)
        )

    def __str__(self):
        # type: () -> str
        return "{project_name} {version}".format(
            project_name=self.project_name, version=self.version
        )


@attr.s(frozen=True)
class Fingerprint(object):
    @classmethod
    def from_stream(
        cls,
        stream,  # type: BinaryIO
        algorithm="sha256",  # type: str
    ):
        # type: (...) -> Fingerprint
        digest = hashlib.new(algorithm)
        hashing.update_hash(filelike=stream, digest=digest)
        return cls(algorithm=algorithm, hash=digest.hexdigest())

    @classmethod
    def from_digest(cls, digest):
        # type: (HashlibHasher) -> Fingerprint
        return cls.from_hashing_fingerprint(digest.hexdigest())

    @classmethod
    def from_hashing_fingerprint(cls, fingerprint):
        # type: (hashing.Fingerprint) -> Fingerprint
        return cls(algorithm=fingerprint.algorithm, hash=fingerprint)

    algorithm = attr.ib()  # type: str
    hash = attr.ib()  # type: str


@attr.s(frozen=True)
class ArtifactURL(object):
    # These ranks prefer the highest digest size and then use alphabetic order for a tie-break.

    _RANKED_ALGORITHMS = tuple(
        sorted(
            hashlib.algorithms_guaranteed,
            key=lambda alg: (-hashlib.new(alg).digest_size, alg),
        )
    )

    @classmethod
    def parse(cls, url):
        # type: (str) -> ArtifactURL
        url_info = urlparse.urlparse(url)
        scheme = parse_scheme(url_info.scheme) if url_info.scheme else None
        path = url_unquote(url_info.path)

        fingerprints = []
        fragment_parameters = urlparse.parse_qs(url_info.fragment)
        if fragment_parameters:
            # Artifact URLs from indexes may contain pre-computed hashes. We isolate those here,
            # centrally, if present.
            # See: https://peps.python.org/pep-0503/#specification
            for alg in cls._RANKED_ALGORITHMS:
                hashes = fragment_parameters.pop(alg, None)
                if not hashes:
                    continue
                if len(hashes) > 1 and len(set(hashes)) > 1:
                    TRACER.log(
                        "The artifact url contains multiple distinct hash values for the {alg} "
                        "algorithm, not trusting any of these: {url}".format(alg=alg, url=url)
                    )
                    continue
                fingerprints.append(Fingerprint(algorithm=alg, hash=hashes[0]))

        download_url = urlparse.urlunparse(
            url_info._replace(
                fragment="&".join(
                    sorted(
                        "{name}={value}".format(name=name, value=value)
                        for name, values in fragment_parameters.items()
                        for value in values
                    )
                )
            )
        )
        normalized_url = urlparse.urlunparse(
            url_info._replace(path=path, params="", query="", fragment="")
        )
        return cls(
            raw_url=url,
            download_url=download_url,
            normalized_url=normalized_url,
            scheme=scheme,
            path=path,
            fragment_parameters=fragment_parameters,
            fingerprints=tuple(fingerprints),
        )

    raw_url = attr.ib(eq=False)  # type: str
    download_url = attr.ib(eq=False)  # type: str
    normalized_url = attr.ib()  # type: str
    scheme = attr.ib(eq=False)  # type: Optional[Union[str, ArchiveScheme.Value, VCSScheme]]
    path = attr.ib(eq=False)  # type: str
    fragment_parameters = attr.ib(eq=False)  # type: Mapping[str, Sequence[str]]
    fingerprints = attr.ib(eq=False)  # type: Tuple[Fingerprint, ...]

    @property
    def is_wheel(self):
        # type: () -> bool
        return self.path.endswith(".whl")

    @property
    def fingerprint(self):
        # type: () -> Optional[Fingerprint]
        return self.fingerprints[0] if self.fingerprints else None


def _convert_url(value):
    # type: (Union[str, ArtifactURL]) -> ArtifactURL
    if isinstance(value, ArtifactURL):
        return value
    return ArtifactURL.parse(value)


@attr.s(frozen=True)
class PartialArtifact(object):
    url = attr.ib(converter=_convert_url)  # type: ArtifactURL
    fingerprint = attr.ib(default=None)  # type: Optional[Fingerprint]
    verified = attr.ib(default=False)  # type: bool


@attr.s(frozen=True)
class ResolvedRequirement(object):
    pin = attr.ib()  # type: Pin
    artifact = attr.ib()  # type: PartialArtifact
    additional_artifacts = attr.ib(default=())  # type: Tuple[PartialArtifact, ...]

    def iter_artifacts(self):
        # type: () -> Iterator[PartialArtifact]
        yield self.artifact
        for artifact in self.additional_artifacts:
            yield artifact

    def iter_artifacts_to_fingerprint(self):
        # type: () -> Iterator[PartialArtifact]
        for artifact in self.iter_artifacts():
            if not artifact.fingerprint:
                yield artifact
